###
# The installation hook mechanism has been greatly inspired by allure-behave
# (see: https://github.com/allure-framework/allure-python/blob/master/allure-behave/src/hooks.py)
###

import os
import inspect

from slugify import slugify
import six

from lemoncheesecake.reporting import ReportLocation
from lemoncheesecake.reporting.savingstrategy import make_report_saving_strategy, DEFAULT_REPORT_SAVING_STRATEGY
from lemoncheesecake.reporting.reportdir import create_report_dir_with_rotation
from lemoncheesecake.reporting.backend import get_reporting_backend_names, parse_reporting_backend_names_expression, \
    get_reporting_backends_for_test_run, get_reporting_backends
from lemoncheesecake.session import Session
from lemoncheesecake.suite import Test, Suite
from lemoncheesecake.events import SyncEventManager


_HOOK_NAMES = (
    "before_all", "after_all",
    "before_feature", "after_feature",
    "before_scenario", "after_scenario",
    "before_step", "after_step"
)

_DEFAULT_REPORTING_BACKENDS = "json", "html"


def _wrap_hook(orig, new):
    def hook(*args, **kwargs):
        new(*args, **kwargs)
        return orig(*args, **kwargs)
    return hook


def install_hooks():
    frame_info = inspect.stack()[1]
    namespace = frame_info[0].f_locals
    hooks = _Hooks(os.path.dirname(frame_info[1]))
    for hook_name in _HOOK_NAMES:
        if hook_name in namespace:
            namespace[hook_name] = _wrap_hook(namespace[hook_name], getattr(hooks, hook_name))
        else:
            namespace[hook_name] = getattr(hooks, hook_name)


def _init_reporting_session(top_dir):
    report_dir = os.environ.get("LCC_REPORT_DIR")
    if report_dir:
        os.mkdir(report_dir)
    else:
        report_dir = create_report_dir_with_rotation(top_dir)

    report_saving_strategy = make_report_saving_strategy(
        os.environ.get("LCC_SAVE_REPORT", DEFAULT_REPORT_SAVING_STRATEGY)
    )

    if "LCC_REPORTING" in os.environ:
        try:
            reporting_backend_names = get_reporting_backend_names(
                _DEFAULT_REPORTING_BACKENDS,
                parse_reporting_backend_names_expression(os.environ["LCC_REPORTING"])
            )
        except ValueError as e:
            raise Exception("Invalid $LCC_REPORTING: %s" % e)
    else:
        reporting_backend_names = _DEFAULT_REPORTING_BACKENDS

    reporting_backends = get_reporting_backends_for_test_run(
        {b.get_name(): b for b in get_reporting_backends()}, reporting_backend_names
    )

    return Session.create(SyncEventManager.load(), reporting_backends, report_dir, report_saving_strategy)


class _Hooks(object):
    def __init__(self, top_dir):
        self.top_dir = top_dir
        self.session = None

    @staticmethod
    def _make_suite_description(feature):
        if feature.description:
            return "Feature: %s\n%s" % (feature.name, "\n".join(feature.description))
        else:
            return "Feature: %s" % feature.name

    @staticmethod
    def _make_suite_name(feature):
        return slugify(feature.name, separator="_")

    @staticmethod
    def _make_test_description(scenario):
        # scenario outlines end with a blank character... strip it:
        scenario_name = scenario.name.strip()
        if scenario.description:
            return "Scenario: %s\n%s" % (scenario_name, "\n".join(scenario.description))
        else:
            return "Scenario: %s" % scenario_name

    @staticmethod
    def _make_test_name(scenario):
        return slugify(scenario.name, separator="_")

    @staticmethod
    def _make_suite_from_feature(feature):
        suite = Suite(
            None, _Hooks._make_suite_name(feature), _Hooks._make_suite_description(feature)
        )
        suite.tags.extend(map(six.text_type, feature.tags))
        return suite

    @staticmethod
    def _make_test_from_scenario(scenario, context):
        test = Test(
            _Hooks._make_test_name(scenario), _Hooks._make_test_description(scenario), None
        )
        test.parent_suite = context.current_suite
        test.tags.extend(map(six.text_type, scenario.tags))
        return test

    def before_all(self, _):
        self.session = _init_reporting_session(self.top_dir)
        self.session.start_test_session()

    def after_all(self, _):
        self.session.end_test_session()

    def before_feature(self, context, feature):
        context.current_suite = _Hooks._make_suite_from_feature(feature)
        self.session.start_suite(context.current_suite)

    def after_feature(self, context, _):
        self.session.end_suite(context.current_suite)

    def before_scenario(self, context, scenario):
        context.current_test = _Hooks._make_test_from_scenario(scenario, context)
        self.session.start_test(context.current_test)

    def after_scenario(self, context, _):
        self.session.end_test(context.current_test)

    def before_step(self, _, step):
        self.session.set_step(step.keyword + " " + step.name)

    def after_step(self, context, step):
        if not self.session.is_successful(ReportLocation.in_test(context.current_test)):
            step.hook_failed = True



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os.path
from contextlib import contextmanager
import shutil
import threading
import traceback
from typing import Optional
import warnings
import functools

import six

from lemoncheesecake.reporting import Report, ReportWriter, ReportLocation, Log
from lemoncheesecake import events
from lemoncheesecake.helpers.typecheck import check_type_string, check_type_bool
from lemoncheesecake.exceptions import AbortTest

_ATTACHMENTS_DIR = "attachments"


def _get_thread_id():
    return threading.current_thread().ident


class _Cursor(object):
    def __init__(self, location, step=None):
        self.location = location
        self.step = step
        self.pending_events = []


class Session(object):
    _instance = None

    def __init__(self, event_manager, report_dir, report):
        self.event_manager = event_manager
        self.report_dir = report_dir
        self.report = report
        self.aborted = False
        self._attachments_dir = os.path.join(self.report_dir, _ATTACHMENTS_DIR)
        self._attachment_count = 0
        self._attachment_lock = threading.Lock()
        self._failures = set()
        self._local = threading.local()

    @classmethod
    def create(cls, event_manager, reporting_backends, report_dir, report_saving_strategy,
               nb_threads=1, parallelized=None):
        report = Report()
        report.nb_threads = nb_threads
        event_manager.add_listener(ReportWriter(report))

        cls._instance = cls(event_manager, report_dir, report)

        # hint: tests with nb_threads > 1 are not actually parallelized if there is only one test
        # that's why there is a dedicated parallelized argument alongside nb_threads
        if parallelized is None:
            parallelized = nb_threads > 1

        for backend in reporting_backends:
            event_manager.add_listener(
                backend.create_reporting_session(report_dir, report, parallelized, report_saving_strategy)
            )

        return cls._instance

    @classmethod
    def get(cls):
        assert cls._instance, "Session has not been created (yet)"
        return cls._instance

    @property
    def cursor(self):  # type: () -> _Cursor
        return self._local.cursor

    @cursor.setter
    def cursor(self, cursor):
        self._local.cursor = cursor

    def _hold_event(self, event):
        self.cursor.pending_events.append(event)

    def _flush_pending_events(self):
        for event in self.cursor.pending_events:
            self.event_manager.fire(event)
        del self.cursor.pending_events[:]

    def _discard_pending_event_if_any(self, event_class):
        if self.cursor.pending_events and isinstance(self.cursor.pending_events[-1], event_class):
            self.cursor.pending_events.pop()
            return True
        else:
            return False

    def _discard_or_fire_event(self, event_class, event):
        discarded = self._discard_pending_event_if_any(event_class)
        if not discarded:
            self.event_manager.fire(event)

    def _mark_location_as_failed(self, location):
        self._failures.add(location)

    def is_successful(self, location=None):
        if location:
            return location not in self._failures
        else:
            return len(self._failures) == 0

    def start_step(self, description):
        self._end_step_if_any()
        self.cursor.step = description
        self._hold_event(
            events.StepStartEvent(self.cursor.location, description, _get_thread_id())
        )
    set_step = start_step

    def end_step(self):
        assert self.cursor.step, "There is no started step"
        self._discard_or_fire_event(
            events.StepStartEvent, events.StepEndEvent(self.cursor.location, self.cursor.step, _get_thread_id())
        )
        self.cursor.step = None

    def _end_step_if_any(self):
        if self.cursor.step:
            self.end_step()

    def _log(self, level, content):
        self._flush_pending_events()
        if level == Log.LEVEL_ERROR:
            self._mark_location_as_failed(self.cursor.location)
        self.event_manager.fire(
            events.LogEvent(self.cursor.location, self.cursor.step, _get_thread_id(), level, content)
        )

    def log_debug(self, content):
        return self._log(Log.LEVEL_DEBUG, content)

    def log_info(self, content):
        return self._log(Log.LEVEL_INFO, content)

    def log_warning(self, content):
        return self._log(Log.LEVEL_WARN, content)

    def log_error(self, content):
        return self._log(Log.LEVEL_ERROR, content)

    def log_check(self, description, is_successful, details):
        self._flush_pending_events()
        if is_successful is False:
            self._mark_location_as_failed(self.cursor.location)
        self.event_manager.fire(events.CheckEvent(
            self.cursor.location, self.cursor.step, _get_thread_id(), description, is_successful, details
        ))

    def log_url(self, url, description):
        self._flush_pending_events()
        self.event_manager.fire(
            events.LogUrlEvent(self.cursor.location, self.cursor.step, _get_thread_id(), url, description)
        )

    @contextmanager
    def prepare_attachment(self, filename, description, as_image=False):
        with self._attachment_lock:
            attachment_filename = "%04d_%s" % (self._attachment_count + 1, filename)
            self._attachment_count += 1
            if not os.path.exists(self._attachments_dir):
                os.mkdir(self._attachments_dir)

        yield os.path.join(self._attachments_dir, attachment_filename)

        self._flush_pending_events()
        self.event_manager.fire(events.LogAttachmentEvent(
            self.cursor.location, self.cursor.step, _get_thread_id(),
            "%s/%s" % (_ATTACHMENTS_DIR, attachment_filename), description, as_image
        ))

    def start_test_session(self):
        self.event_manager.fire(events.TestSessionStartEvent(self.report))

    def end_test_session(self):
        self.event_manager.fire(events.TestSessionEndEvent(self.report))

    def start_test_session_setup(self):
        self.cursor = _Cursor(ReportLocation.in_test_session_setup())
        self._hold_event(events.TestSessionSetupStartEvent())

    def end_test_session_setup(self):
        self._end_step_if_any()
        self._discard_or_fire_event(events.TestSessionSetupStartEvent, events.TestSessionSetupEndEvent())

    def start_test_session_teardown(self):
        self.cursor = _Cursor(ReportLocation.in_test_session_teardown())
        self._hold_event(events.TestSessionTeardownStartEvent())

    def end_test_session_teardown(self):
        self._end_step_if_any()
        self._discard_or_fire_event(events.TestSessionTeardownStartEvent, events.TestSessionTeardownEndEvent())

    def start_suite(self, suite):
        self.event_manager.fire(events.SuiteStartEvent(suite))

    def end_suite(self, suite):
        self.event_manager.fire(events.SuiteEndEvent(suite))

    def start_suite_setup(self, suite):
        self.cursor = _Cursor(ReportLocation.in_suite_setup(suite))
        self._hold_event(events.SuiteSetupStartEvent(suite))

    def end_suite_setup(self, suite):
        self._end_step_if_any()
        self._discard_or_fire_event(events.SuiteSetupStartEvent, events.SuiteSetupEndEvent(suite))

    def start_suite_teardown(self, suite):
        self.cursor = _Cursor(ReportLocation.in_suite_teardown(suite))
        self._hold_event(events.SuiteTeardownStartEvent(suite))

    def end_suite_teardown(self, suite):
        self._end_step_if_any()
        self._discard_or_fire_event(events.SuiteTeardownStartEvent, events.SuiteTeardownEndEvent(suite))

    def start_test(self, test):
        self.event_manager.fire(events.TestStartEvent(test))
        self.cursor = _Cursor(ReportLocation.in_test(test))

    def end_test(self, test):
        self._end_step_if_any()
        self.event_manager.fire(events.TestEndEvent(test))

    def skip_test(self, test, reason):
        self.event_manager.fire(events.TestSkippedEvent(test, reason))
        self._mark_location_as_failed(ReportLocation.in_test(test))

    def disable_test(self, test, reason):
        self.event_manager.fire(events.TestDisabledEvent(test, reason))


def _interruptible(wrapped):
    @functools.wraps(wrapped)
    def wrapper(*args, **kwargs):
        if Session.get().aborted:
            raise AbortTest("tests have been manually stopped")
        return wrapped(*args, **kwargs)
    wrapper.__doc__ = wrapped.__doc__
    return wrapper


@_interruptible
def set_step(description, detached=NotImplemented):
    # type: (str, bool) -> None
    """
    Set a new step.

    :param description: the step description
    :param detached: argument deprecated since 1.4.5, does nothing.
    """
    if detached is not NotImplemented:
        warnings.warn(
            "The 'detached' argument does no longer do anything (deprecated since version 1.4.5)",
            DeprecationWarning, stacklevel=2
        )

    check_type_string("description", description)
    Session.get().start_step(description)


def end_step(step):
    """
    Function deprecated since version 1.4.5, does nothing.
    """
    warnings.warn(
        "The 'end_step' function is deprecated since version 1.4.5, it actually does nothing.",
        DeprecationWarning, stacklevel=2
    )


@contextmanager
@_interruptible
def detached_step(description):
    """
    Context manager deprecated since version 1.4.5, it only does a set_step.
    """
    warnings.warn(
        "The 'detached_step' context manager is deprecated since version 1.4.5. Use 'set_step' function instead.",
        DeprecationWarning, stacklevel=2
    )
    set_step(description)
    yield


@_interruptible
def log_debug(content):
    # type: (str) -> None
    """
    Log a debug level message.
    """
    check_type_string("content", content)
    Session.get().log_debug(content)


@_interruptible
def log_info(content):
    # type: (str) -> None
    """
    Log a info level message.
    """
    check_type_string("content", content)
    Session.get().log_info(content)


@_interruptible
def log_warning(content):
    # type: (str) -> None
    """
    Log a warning level message.
    """
    check_type_string("content", content)
    Session.get().log_warning(content)


@_interruptible
def log_error(content):
    # type: (str) -> None
    """
    Log an error level message.
    """
    check_type_string("content", content)
    Session.get().log_error(content)


@_interruptible
def log_check(description, is_successful, details=None):
    check_type_string("description", description)
    check_type_bool("is_successful", is_successful)
    check_type_string("details", details, optional=True)

    Session.get().log_check(description, is_successful, details)


def _prepare_attachment(filename, description=None, as_image=False):
    check_type_string("description", description, optional=True)
    check_type_bool("as_image", as_image)

    return Session.get().prepare_attachment(filename, description or filename, as_image=as_image)


@_interruptible
def prepare_attachment(filename, description=None):
    """
    Context manager. Prepare a attachment using a pseudo filename and an optional description.
    It returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description)


@_interruptible
def prepare_image_attachment(filename, description=None):
    """
    Context manager. Prepare an image attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description, as_image=True)


def _save_attachment_file(filename, description=None, as_image=False):
    with _prepare_attachment(os.path.basename(filename), description, as_image=as_image) as report_attachment_path:
        shutil.copy(filename, report_attachment_path)


@_interruptible
def save_attachment_file(filename, description=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description)


@_interruptible
def save_image_file(filename, description=None):
    """
    Save an image using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description, as_image=True)


def _save_attachment_content(content, filename, description=None, as_image=False):
    if type(content) is six.text_type:
        opening_mode = "w"
        if six.PY2:
            content = content.encode("utf-8")
    else:
        opening_mode = "wb"

    with _prepare_attachment(filename, description, as_image=as_image) as path:
        with open(path, opening_mode) as fh:
            fh.write(content)


@_interruptible
def save_attachment_content(content, filename, description=None):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, as_image=False)


@_interruptible
def save_image_content(content, filename, description=None):
    """
    Save a given image content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, as_image=True)


@_interruptible
def log_url(url, description=None):
    # type: (str, Optional[str]) -> None
    """
    Log an URL.
    """
    check_type_string("url", url)
    check_type_string("description", description, optional=True)

    Session.get().log_url(url, description or url)


def add_report_info(name, value):
    # type: (str, str) -> None

    check_type_string("name", name)
    check_type_string("value", value)

    session = Session.get()
    session.report.add_info(name, value)


class Thread(threading.Thread):
    """
    Acts exactly as the standard threading.Thread class and must be used instead when running threads
    within a test.
    """
    def __init__(self, *args, **kwargs):
        super(Thread, self).__init__(*args, **kwargs)
        self._session = Session.get()

        # flush result starting event if any
        cursor = self._session.cursor
        if self._session.cursor.pending_events and not isinstance(cursor.pending_events[0], events.StepStartEvent):
            event = cursor.pending_events.pop(0)
            self._session.event_manager.fire(event)

        # keep track of the current location and step
        self._cursor = _Cursor(cursor.location)
        self._default_step = self._session.cursor.step

    def run(self):
        self._session.cursor = self._cursor
        self._session.start_step(self._default_step)
        try:
            return super(Thread, self).run()
        except Exception:
            # FIXME: use exception instead of last implicit stacktrace
            stacktrace = traceback.format_exc()
            if six.PY2:
                stacktrace = stacktrace.decode("utf-8", "replace")
            log_error("Caught unexpected exception while running test: " + stacktrace)
        finally:
            self._session.end_step()



def humanize_duration(duration, show_milliseconds=False):
    ret = ""

    if duration / 3600 >= 1:
        ret += ("%02dh" if ret else "%dh") % (duration / 3600)
        duration %= 3600

    if duration / 60 >= 1:
        ret += ("%02dm" if ret else "%dm") % (duration / 60)
        duration %= 60

    if show_milliseconds:
        if duration >= 0:
            ret += ("%06.03fs" if ret else "%.03fs") % duration
    else:
        if duration >= 1:
            ret += ("%02ds" if ret else "%ds") % duration
        if ret == "":
            ret = "%.03fs" % duration

    return ret



from functools import reduce

from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.helpers.text import ensure_single_line_text
from lemoncheesecake.helpers.console import print_table
from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends, add_report_path_cli_arg, get_report_path
from lemoncheesecake.reporting import load_report
from lemoncheesecake.filter import add_result_filter_cli_args, make_result_filter, \
    add_step_filter_cli_args, make_step_filter
from lemoncheesecake.testtree import flatten_suites, filter_suites


def get_total_duration(elems):
    return reduce(lambda x, y: x + y, (elem.duration or 0 for elem in elems), 0)


class TopTests(Command):
    def get_name(self):
        return "top-tests"

    def get_description(self):
        return "Display tests ordered by duration"

    def add_cli_args(self, cli_parser):
        add_result_filter_cli_args(cli_parser, only_executed_tests=True)
        group = cli_parser.add_argument_group("Top tests")
        add_report_path_cli_arg(group)

    @staticmethod
    def _get_tests_ordered_by_duration(tests):
        return sorted(tests, key=lambda test: test.duration, reverse=True)

    @staticmethod
    def _format_test_entry(test, total_time):
        if test.duration is not None:
            return (
                test.path,
                humanize_duration(test.duration, show_milliseconds=True),
                "%d%%" % (test.duration / total_time * 100) if total_time else 100
            )
        else:
            return test.path, "-", "-"

    @staticmethod
    def get_top_tests(report, test_filter):
        tests = TopTests._get_tests_ordered_by_duration(filter(test_filter, report.all_tests()))
        total_duration = get_total_duration(tests)
        return [TopTests._format_test_entry(test, total_duration) for test in tests]

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        test_filter = make_result_filter(cli_args, only_executed_tests=True)

        print_table(
            "Tests, ordered by duration",
            ("Test", "Duration", "In %"),
            TopTests.get_top_tests(report, test_filter)
        )

        return 0


class TopSuites(Command):
    def get_name(self):
        return "top-suites"

    def get_description(self):
        return "Display suites ordered by duration"

    def add_cli_args(self, cli_parser):
        add_result_filter_cli_args(cli_parser, only_executed_tests=True)
        group = cli_parser.add_argument_group("Top suites")
        add_report_path_cli_arg(group)

    @staticmethod
    def _get_suites_ordered_by_duration(suites):
        return sorted(
            filter(lambda s: any((s.get_tests(), s.suite_setup, s.suite_teardown)), suites),
            key=lambda suite: suite.duration,
            reverse=True
        )

    @staticmethod
    def _format_suite_entry(suite, total_time):
        if suite.duration is not None:
            return (
                suite.path,
                len(suite.get_tests()),
                humanize_duration(suite.duration, show_milliseconds=True),
                "%d%%" % ((suite.duration / total_time * 100) if total_time else 100)
            )
        else:
            return suite.path, len(suite.get_tests()), "-", "-"

    @staticmethod
    def get_top_suites(report, result_filter):
        processed_suites = TopSuites._get_suites_ordered_by_duration(
            flatten_suites(filter_suites(report.get_suites(), result_filter))
        )
        total_duration = get_total_duration(processed_suites)
        return [TopSuites._format_suite_entry(suite, total_duration) for suite in processed_suites]

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        result_filter = make_result_filter(cli_args, only_executed_tests=True)

        print_table(
            "Suites, ordered by duration",
            ("Suite", "Tests Nb.", "Duration", "In %"),
            TopSuites.get_top_suites(report, result_filter)
        )

        return 0


class TopSteps(Command):
    def get_name(self):
        return "top-steps"

    def get_description(self):
        return "Display steps aggregated and ordered by duration"

    def add_cli_args(self, cli_parser):
        add_step_filter_cli_args(cli_parser)
        group = cli_parser.add_argument_group("Top steps")
        add_report_path_cli_arg(group)

    @staticmethod
    def _group_steps_by_description(steps):
        steps_by_description = {}
        for step in steps:
            description = ensure_single_line_text(step.description)
            if description not in steps_by_description:
                steps_by_description[description] = []
            steps_by_description[description].append(step)
        return steps_by_description

    @staticmethod
    def _get_steps_min_duration(steps):
        return min(step.duration or 0 for step in steps)

    @staticmethod
    def _get_steps_max_duration(steps):
        return max(step.duration or 0 for step in steps)

    @staticmethod
    def _get_steps_average_duration(steps):
        return get_total_duration(steps) / len(steps)

    @staticmethod
    def _format_steps_aggregation(description, occurrences, min_duration, max_duration, average_duration, duration,
                                  duration_pct):
        return (
            description,
            str(occurrences),
            humanize_duration(min_duration, show_milliseconds=True),
            humanize_duration(max_duration, show_milliseconds=True),
            humanize_duration(average_duration, show_milliseconds=True),
            humanize_duration(duration, show_milliseconds=True),
            "%d%%" % duration_pct
        )

    @staticmethod
    def _get_steps_aggregation(steps):
        steps_by_description = TopSteps._group_steps_by_description(steps)
        total_duration = get_total_duration(steps)

        data = []
        for description, steps in steps_by_description.items():
            duration = get_total_duration(steps)
            data.append([
                description,
                len(steps),
                TopSteps._get_steps_min_duration(steps),
                TopSteps._get_steps_max_duration(steps),
                TopSteps._get_steps_average_duration(steps),
                duration,
                (duration / total_duration * 100) if total_duration else 100
            ])

        return sorted(data, key=lambda row: row[-2], reverse=True)

    @staticmethod
    def get_top_steps(report, step_filter):
        steps = list(filter(step_filter, report.all_steps()))
        steps_aggregation = TopSteps._get_steps_aggregation(steps)
        return [TopSteps._format_steps_aggregation(*agg) for agg in steps_aggregation]

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)

        report = load_report(report_path, auto_detect_reporting_backends())
        step_filter = make_step_filter(cli_args)

        print_table(
            "Steps, aggregated and ordered by duration",
            ("Step", "Occ.", "Min.", "Max", "Avg.", "Total", "In %"),
            TopSteps.get_top_steps(report, step_filter)
        )

        return 0



'''
Created on Apr 3, 2017

@author: nicolas
'''

import six
from typing import List, Any

from lemoncheesecake.helpers.text import STRING_TYPES, jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.value import is_

_TYPE_NAMES = {
    int: "integer",
    float: "float",
    str: "string",
    dict: "collection",
    list: "array", tuple: "array"
}
if six.PY2:
    _TYPE_NAMES[unicode] = "string",


class IsValueOfType(Matcher):
    def __init__(self, types, type_name, value_matcher):
        # type: (List[Any], str, Matcher) -> None
        self.types = types
        self.type_name = type_name
        self.value_matcher = value_matcher

    def build_description(self, transformation):
        ret = transformation("to be %s" % self.type_name)
        if self.value_matcher:
            ret += " that %s" % self.value_matcher.build_description(MatcherDescriptionTransformer(conjugate=True))

        return ret

    @staticmethod
    def _get_value_type_name(value):
        try:
            return _TYPE_NAMES[type(value)]
        except KeyError:
            return str(type(value))

    def matches(self, actual):
        if type(actual) in self.types:
            if self.value_matcher:
                return self.value_matcher.matches(actual)
            else:
                return MatchResult.success("got %s" % jsonify(actual))
        else:
            return MatchResult.failure("got %s (%s)" % (jsonify(actual), self._get_value_type_name(actual)))


def _is_type(types, type_name):
    def wrapper(value_matcher=None):
        return IsValueOfType(
            types, type_name, is_(value_matcher) if value_matcher is not None else None
        )
    wrapper.__doc__ = "Test if value is of type %s" % type_name
    return wrapper


is_integer = _is_type([int], "an integer")
is_float = _is_type([float], "a float")
is_bool = _is_type([bool], "a boolean")
is_str = _is_type(STRING_TYPES, "a string")
is_dict = _is_type([dict], "a collection")
is_list = _is_type([list, tuple], "a list")



from .console import ConsoleBackend
from .xml import XmlBackend
from .json_ import JsonBackend
from .html import HtmlBackend
from .junit import JunitBackend
from .reportportal import ReportPortalBackend
from .slack import SlackReportingBackend


# NB: order matters (a bit), we typically want the messages of HtmlBackend to appear
# after those from ConsoleBackend we running "lcc run"
REPORTING_BACKENDS = ConsoleBackend, XmlBackend, JsonBackend, HtmlBackend, JunitBackend, \
    ReportPortalBackend, SlackReportingBackend



'''
Created on Mar 29, 2016

@author: nicolas
'''

import os

from lemoncheesecake.helpers.orderedset import OrderedSet
from lemoncheesecake.exceptions import LemoncheesecakeException

_NEGATION_FLAGS = "-^~"


class ReportingSession(object):
    pass


class ReportingSessionBuilderMixin(object):
    def create_reporting_session(self, report_dir, report, parallel, report_saving_strategy):
        raise NotImplementedError()


class ReportSerializerMixin(object):
    def save_report(self, filename, report):
        raise NotImplementedError()


class ReportUnserializerMixin(object):
    def load_report(self, filename):
        raise NotImplementedError()


class ReportingBackend(object):
    def get_name(self):
        raise NotImplementedError()

    def is_available(self):
        return True


class FileReportSession(ReportingSession):
    def __init__(self, report_filename, report, reporting_backend, report_saving_strategy):
        self.report_filename = report_filename
        self.report = report
        self.reporting_backend = reporting_backend
        self.report_saving_strategy = report_saving_strategy

    def _save(self):
        self.reporting_backend.save_report(self.report_filename, self.report)

    def _handle_event(self, event):
        report_must_be_saved = self.report_saving_strategy and self.report_saving_strategy(event, self.report)
        if report_must_be_saved:
            self._save()

    on_test_session_setup_end = _handle_event
    on_test_session_teardown_end = _handle_event
    on_suite_setup_end = _handle_event
    on_suite_teardown_end = _handle_event
    on_test_end = _handle_event
    on_suite_end = _handle_event
    on_log = _handle_event
    on_log_attachment = _handle_event
    on_log_url = _handle_event
    on_check = _handle_event

    def on_test_session_end(self, event):
        # no matter what is the report_saving_strategy,
        # the report will always be saved at the end of tests
        self._save()


class FileReportBackend(ReportingBackend, ReportSerializerMixin, ReportingSessionBuilderMixin):
    def get_report_filename(self):
        raise NotImplementedError()

    def create_reporting_session(self, report_dir, report, parallel, report_saving_strategy):
        return FileReportSession(
            os.path.join(report_dir, self.get_report_filename()), report, self, report_saving_strategy
        )


def filter_available_reporting_backends(backends):
    return list(filter(lambda backend: backend.is_available(), backends))


def get_reporting_backends():
    from lemoncheesecake.reporting.backends import REPORTING_BACKENDS
    return list(
        filter(
            lambda backend: backend.is_available(),
            (backend_class() for backend_class in REPORTING_BACKENDS)
        )
    )


def get_reporting_backend_names(default_names, custom_names):
    if all(name[0] not in ("+" + _NEGATION_FLAGS) for name in custom_names):  # fixed list
        return custom_names

    elif all(name[0] in ("+" + _NEGATION_FLAGS) for name in custom_names):  # turn on/off directives
        names = OrderedSet(default_names)
        for specified_name in custom_names:
            if specified_name[0] == "+":  # turn on
                names.add(specified_name[1:])
            else:  # turn off
                try:
                    names.remove(specified_name[1:])
                except KeyError:
                    raise ValueError(
                        "reporting backend '%s' is not among the default reporting backends" % specified_name[1:]
                    )
        return names

    else:
        raise ValueError(
            "either the custom reporting backends must be fixed backend list, "
            "or a list of turn on/off (+ / ^) directives"
        )


def parse_reporting_backend_names_expression(expr):
    return list(filter(bool, expr.split(" ")))


def get_reporting_backend_by_name(name):
    try:
        return next(backend for backend in get_reporting_backends() if backend.get_name() == name)
    except StopIteration:
        raise KeyError()


def get_reporting_backends_for_test_run(available_backends, backend_names):
    backends = []
    for backend_name in backend_names:
        try:
            backend = available_backends[backend_name]
        except KeyError:
            raise LemoncheesecakeException("Unknown reporting backend '%s'" % backend_name)
        if not isinstance(backend, ReportingSessionBuilderMixin):
            raise LemoncheesecakeException("Reporting backend '%s' is not suitable for test run" % backend_name)
        backends.append(backend)
    return backends



import re
import textwrap
import itertools
import json

import six


STRING_TYPES = [str] if six.PY3 else [str, unicode]


# borrowed from https://stackoverflow.com/a/1176023
def camel_case_to_snake_case(name):
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def ensure_single_line_text(text):
    return text.replace("\n", " ")


def wrap_text(text, width):
    if text:
        return "\n".join(
            itertools.chain.from_iterable(
                textwrap.wrap(line, width) for line in text.split("\n")
            )
        )
    else:
        return ""


def jsonify(data):
    return json.dumps(data, ensure_ascii=False)



from __future__ import print_function

from terminaltables import AsciiTable
from termcolor import colored
import six

from lemoncheesecake.helpers.text import wrap_text
from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.helpers import terminalsize
from lemoncheesecake.filter import ResultFilter
from lemoncheesecake.reporting import Log, Check, Url, Attachment, TestResult


def test_status_to_color(status):
    return {
        "passed": "green",
        "failed": "red",
        "disabled": "cyan",
        "in_progress": "cyan",
        None: "cyan"  # None means test is in progress
    }.get(status, "yellow")


def log_level_to_color(level):
    return {
        "debug": "cyan",
        "info": "cyan",
        "warn": "yellow",
        "error": "red"
    }.get(level, "cyan")


def outcome_to_color(outcome):
    return "green" if outcome else "red"


def _serialize_metadata(tags, properties, links, disabled):
    parts = []

    if disabled:
        parts.append("DISABLED")
    parts.extend(tags)
    parts.extend("%s:%s" % (key, properties[key]) for key in sorted(properties))
    parts.extend(link_name or link_url for link_url, link_name in links)

    return ", ".join(parts)


def serialize_metadata(obj, hide_disabled=False):
    return _serialize_metadata(
        obj.tags, obj.properties, obj.links,
        disabled=not hide_disabled and getattr(obj, "disabled", False)
    )


def serialize_hierarchy_metadata(obj, hide_disabled=False):
    return _serialize_metadata(
        obj.hierarchy_tags, obj.hierarchy_properties, obj.hierarchy_links,
        disabled=not hide_disabled and (hasattr(obj, "is_disabled") and obj.is_disabled())
    )


class Renderer(object):
    def __init__(self, max_width, explicit=False, highlight=None, show_debug_logs=False):
        self.max_width = max_width
        self.explicit = explicit
        self.highlight = highlight
        self.show_debug_logs = show_debug_logs
        # "20" is an approximation of the maximal overhead of table border, padding, and table first cell
        self._table_overhead = 20

    def wrap_description_col(self, description):
        return wrap_text(description, int((self.max_width - self._table_overhead) * 0.75))

    def wrap_details_col(self, details):
        return wrap_text(details, int((self.max_width - self._table_overhead) * 0.25))

    def render_check_outcome(self, is_successful):
        if self.explicit:
            check_label = "CHECK %s" % ("OK" if is_successful else "KO")
        else:
            check_label = "CHECK"

        return colored(check_label, color=outcome_to_color(is_successful), attrs=["bold"])

    def render_highlighted(self, content):
        if not self.highlight or not content:
            return content

        return self.highlight.sub(
            lambda m: colored(m.group(0), color="yellow", attrs=["bold", "underline"]), content
        )

    def render_step_log(self, log):
        if isinstance(log, Log):
            if log.level == "debug" and not self.show_debug_logs:
                return None
            else:
                return [
                    colored(log.level.upper(), color=log_level_to_color(log.level), attrs=["bold"]),
                    self.render_highlighted(self.wrap_description_col(log.message))
                ]
        if isinstance(log, Check):
            return [
                self.render_check_outcome(log.is_successful),
                self.render_highlighted(self.wrap_description_col(log.description)),
                self.render_highlighted(self.wrap_details_col(log.details))
            ]
        if isinstance(log, Url):
            if log.description == log.url:
                description = log.url
            else:
                description = "%s (%s)" % (log.url, log.description)
            return [
                colored("URL", color="cyan", attrs=["bold"]),
                self.render_highlighted(self.wrap_description_col(description))
            ]
        if isinstance(log, Attachment):
            return [
                colored("ATTACH", color="cyan", attrs=["bold"]),
                self.render_highlighted(self.wrap_description_col(log.description)),
                self.render_highlighted(log.filename)
            ]

        raise ValueError("Unknown step log class '%s'" % log.__class__.__name__)

    def render_steps(self, steps):
        rows = []
        for step in steps:
            step_log_rows = list(filter(bool, map(self.render_step_log, step.get_logs())))
            if step_log_rows:
                rows.append([
                    "",
                    colored(
                        self.render_highlighted(self.wrap_description_col(step.description)),
                        color=outcome_to_color(step.is_successful()),
                        attrs=["bold"]
                    ),
                    colored(humanize_duration(step.duration, show_milliseconds=True), attrs=["bold"])
                        if step.duration is not None else "-"
                ])
                rows.extend(step_log_rows)

        if not rows:
            return None

        table = AsciiTable(rows)
        table.inner_heading_row_border = False
        table.justify_columns[0] = "center"
        table.inner_row_border = True

        return table.table

    def render_chunk(self, description, short_description, status, steps):
        if status is None:
            status = "in_progress"

        if steps:
            details = self.render_steps(steps) or "n/a"
        else:
            details = "n/a"

        parts = [
            colored(
                "%s: %s" % (status.upper(), description) if self.explicit else description,
                color=test_status_to_color(status),
                attrs=["bold"]
            )
        ]
        if short_description:
            parts.append(colored("(%s)" % short_description, attrs=["bold"]))
        parts.append(details)

        return "\n".join(parts)

    def render_test(self, test):
        test_metadata = serialize_hierarchy_metadata(test, hide_disabled=True)
        if test_metadata:
            short_description = "%s - %s" % (test.path, test_metadata)
        else:
            short_description = test.path

        return self.render_chunk(test.description, short_description, test.status, test.get_steps())

    def render_result(self, result):
        if result.type == "suite_setup":
            description = "- SUITE SETUP - %s" % result.parent_suite.description
            short_description = result.parent_suite.path
        elif result.type == "suite_teardown":
            description = "- SUITE TEARDOWN - %s" % result.parent_suite.description
            short_description = result.parent_suite.path
        elif result.type == "test_session_setup":
            description = "- TEST SESSION SETUP -"
            short_description = None
        elif result.type == "test_session_teardown":
            description = "- TEST SESSION TEARDOWN -"
            short_description = None
        else:
            raise ValueError("Unknown result type '%s'" % result.type)

        return self.render_chunk(description, short_description, result.status, result.get_steps())

    def render_results(self, results):
        for result in results:
            if isinstance(result, TestResult):
                yield self.render_test(result)
            else:
                yield self.render_result(result)
            yield ""


def _print_chunks(chunks):
    for chunk in chunks:
        print(chunk if six.PY3 else chunk.encode("utf8"))


def print_report(report, result_filter=None, max_width=None, show_debug_logs=False, explicit=False):
    ###
    # Setup terminal
    ###
    if not max_width:
        max_width, _ = terminalsize.get_terminal_size()

    ###
    # Get a generator over data to be printed on the console
    ###
    renderer = Renderer(
        max_width=max_width, show_debug_logs=show_debug_logs, explicit=explicit,
        highlight=result_filter.grep if isinstance(result_filter, ResultFilter) else None
    )
    if not result_filter:
        if report.nb_tests == 0:
            print("No tests found in report")
            return
        chunks = renderer.render_results(report.all_results())
    else:
        results = list(filter(result_filter, report.all_results()))
        if not results:
            print("The filter does not match anything in the report")
            return
        chunks = renderer.render_results(results)

    ###
    # Do the actual job
    ###
    _print_chunks(chunks)



'''
Created on Jan 25, 2017

@author: nicolas
'''

from __future__ import print_function
import argparse
import sys

import colorama

from lemoncheesecake.exceptions import LemoncheesecakeException
from lemoncheesecake.cli.commands import get_commands
from lemoncheesecake.cli.utils import LEMONCHEESECAKE_VERSION


# we implement our own "version" instead of using the built-in argparse "version" action
# because its behavior between Python 2 & 3 is not consistent and because it does
# undesired text-wrapping
class Version(argparse.Action):
    def __call__(self, *dummy_args, **dummy_kwargs):
        print(LEMONCHEESECAKE_VERSION)
        sys.exit(0)


def build_cli_parser():
    cli_parser = argparse.ArgumentParser()
    cli_parser.add_argument("--version", "-v", nargs=0, action=Version)
    cli_sub_parsers = cli_parser.add_subparsers(dest="command")
    cli_sub_parsers.required = True
    for command in get_commands():
        cli_cmd_parser = cli_sub_parsers.add_parser(command.get_name(), help=command.get_description())
        command.add_cli_args(cli_cmd_parser)
    return cli_parser


def build_cli_args(args=None):
    cli_parser = build_cli_parser()
    return cli_parser.parse_args(args)


def main(args=None):
    colorama.init()

    try:
        cli_args = build_cli_args(args)
    except LemoncheesecakeException as excp:
        return str(excp)

    command = next(cmd for cmd in get_commands() if cmd.get_name() == cli_args.command)
    try:
        return command.run_cmd(cli_args)
    except LemoncheesecakeException as excp:
        return str(excp)



'''
Created on Jan 24, 2016

@author: nicolas
'''

import traceback
import itertools

import six

from lemoncheesecake.exceptions import AbortTest, AbortSuite, AbortAllTests, LemoncheesecakeException, \
    UserError, TaskFailure, serialize_current_exception
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.reporting import ReportLocation
from lemoncheesecake.task import BaseTask, TaskContext, run_tasks
from lemoncheesecake.fixture import initialize_fixture_cache


class RunContext(TaskContext):
    def __init__(self, session, fixture_registry, force_disabled, stop_on_failure):
        super(RunContext, self).__init__()
        self.session = session
        self.fixture_registry = fixture_registry
        self.force_disabled = force_disabled
        self.stop_on_failure = stop_on_failure
        self._aborted_session = False
        self._aborted_suites = set()

    def handle_exception(self, excp, suite=None):
        if isinstance(excp, AbortTest):
            self.session.log_error("The test has been aborted: %s" % excp)
        elif isinstance(excp, AbortSuite):
            self.session.log_error("The suite has been aborted: %s" % excp)
            self._aborted_suites.add(suite)
        elif isinstance(excp, AbortAllTests):
            self.session.log_error("All tests have been aborted: %s" % excp)
            self._aborted_session = True
        else:
            # FIXME: use exception instead of last implicit stacktrace
            stacktrace = traceback.format_exc()
            if six.PY2:
                stacktrace = stacktrace.decode("utf-8", "replace")
            self.session.log_error("Caught unexpected exception while running test: " + stacktrace)

    def run_setup_funcs(self, funcs, location):
        teardown_funcs = []
        for setup_func, teardown_func in funcs:
            if setup_func:
                try:
                    setup_func()
                except Exception as e:
                    self.handle_exception(e)
                    break
                else:
                    if not self.session.is_successful(location):
                        break
                    else:
                        teardown_funcs.append(teardown_func)
            else:
                teardown_funcs.append(teardown_func)
        return teardown_funcs

    def run_teardown_funcs(self, teardown_funcs):
        for teardown_func in teardown_funcs:
            if teardown_func:
                try:
                    teardown_func()
                except Exception as e:
                    self.handle_exception(e)

    def enable_task_abort(self):
        super(RunContext, self).enable_task_abort()
        self.session.aborted = True

    def is_task_to_be_skipped(self, task):
        # check for error in base implementation
        skip_reason = super(RunContext, self).is_task_to_be_skipped(task)
        if skip_reason:
            return skip_reason

        # check for error in event handling
        exception, _ = self.session.event_manager.get_pending_failure()
        if exception is not None:
            return str(exception)

        # check for test session abort
        if self._aborted_session:
            return "tests have been aborted"

        # check for suite abort
        if isinstance(task, TestTask):
            if task.test.parent_suite in self._aborted_suites:
                return "the tests of this test suite have been aborted"

        # check for --stop-on-failure
        if self.stop_on_failure and not self.session.is_successful():
            return "tests have been aborted on --stop-on-failure"

        return None


class TestTask(BaseTask):
    def __init__(self, test, suite_scheduled_fixtures, dependency=None):
        BaseTask.__init__(self)
        self.test = test
        self.suite_scheduled_fixtures = suite_scheduled_fixtures
        self.dependencies = [dependency] if dependency else []

    def get_on_success_dependencies(self):
        return self.dependencies

    def _is_test_disabled(self, context):
        return self.test.is_disabled() and not context.force_disabled

    def _handle_disabled_test(self, context):
        disabled = self.test.is_disabled()
        disabled_reason = disabled if isinstance(disabled, six.string_types) else None
        context.session.disable_test(self.test, disabled_reason)

    def skip(self, context, reason=None):
        if self._is_test_disabled(context):
            self._handle_disabled_test(context)
        else:
            context.session.skip_test(self.test, "Test skipped because %s" % reason if reason else None)

    @staticmethod
    def _prepare_test_args(test, scheduled_fixtures):
        args = {}
        for arg_name in test.get_arguments():
            if arg_name in test.parameters:
                args[arg_name] = test.parameters[arg_name]
            else:
                args[arg_name] = scheduled_fixtures.get_fixture_result(arg_name)

        return args

    def run(self, context):
        ###
        # Checker whether the test must be executed or not
        ###
        if self._is_test_disabled(context):
            self._handle_disabled_test(context)
            return

        ###
        # Begin test
        ###
        context.session.start_test(self.test)

        ###
        # Setup test (setup and fixtures)
        ###
        suite = self.test.parent_suite

        if suite.has_hook("setup_test"):
            def setup_test_wrapper():
                suite.get_hook("setup_test")(self.test)
        else:
            setup_test_wrapper = None

        if suite.has_hook("teardown_test"):
            def teardown_test_wrapper():
                status_so_far = "passed" if context.session.is_successful(ReportLocation.in_test(self.test)) else "failed"
                suite.get_hook("teardown_test")(self.test, status_so_far)
        else:
            teardown_test_wrapper = None

        setup_teardown_funcs = list()
        setup_teardown_funcs.append((setup_test_wrapper, teardown_test_wrapper))
        scheduled_fixtures = context.fixture_registry.get_fixtures_scheduled_for_test(
            self.test, self.suite_scheduled_fixtures
        )
        setup_teardown_funcs.extend(scheduled_fixtures.get_setup_teardown_pairs())

        context.session.set_step("Setup test")

        if any(setup for setup, _ in setup_teardown_funcs):
            teardown_funcs = context.run_setup_funcs(setup_teardown_funcs, ReportLocation.in_test(self.test))
        else:
            teardown_funcs = [teardown for _, teardown in setup_teardown_funcs if teardown]

        ###
        # Run test:
        ###
        if context.session.is_successful(ReportLocation.in_test(self.test)):
            test_args = self._prepare_test_args(self.test, scheduled_fixtures)
            context.session.set_step(self.test.description)
            try:
                self.test.callback(**test_args)
            except Exception as e:
                context.handle_exception(e, suite)

        ###
        # Teardown
        ###
        if any(teardown_funcs):
            context.session.set_step("Teardown test")
            context.run_teardown_funcs(teardown_funcs)

        context.session.end_test(self.test)

        if not context.session.is_successful(ReportLocation.in_test(self.test)):
            raise TaskFailure("test '%s' failed" % self.test.path)

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.test.path)


def build_test_task(test, suite_scheduled_fixtures, dependency):
    return TestTask(test, suite_scheduled_fixtures, dependency)


def build_suite_tasks(
        suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task,
        parent_suite_beginning_task=None, force_disabled=False):
    ###
    # Build suite beginning task
    ###
    suite_beginning_task = build_suite_beginning_task(
        suite, list((filter(bool, (test_session_setup_task, parent_suite_beginning_task))))
    )

    ###
    # Build suite setup task (if any)
    ###
    suite_scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_suite(
        suite, session_scheduled_fixtures, force_disabled
    )
    suite_setup_task = build_suite_initialization_task(
        suite, suite_scheduled_fixtures, [suite_beginning_task], force_disabled
    )

    ###
    # Build test tasks
    ###
    test_dependency = suite_setup_task if suite_setup_task else suite_beginning_task
    test_tasks = [
        build_test_task(test, suite_scheduled_fixtures, test_dependency)
        for test in suite.get_tests()
    ]

    ###
    # Build suite teardown task (if any)
    ###
    suite_teardown_task = build_suite_teardown_task(suite, suite_setup_task, test_tasks)

    ###
    # Build sub suite tasks
    ###
    sub_suite_tasks = []
    for sub_suite in suite.get_suites():
        sub_suite_tasks.extend(
            build_suite_tasks(
                sub_suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task, suite_beginning_task,
                force_disabled
            )
        )

    ###
    # Build suite ending task
    ###
    suite_ending_dependencies = []
    suite_ending_dependencies.extend(test_tasks)
    if suite_teardown_task:
        suite_ending_dependencies.append(suite_teardown_task)
    suite_ending_dependencies.extend(
        task for task in sub_suite_tasks if isinstance(task, SuiteEndingTask) and task.suite in suite.get_suites()
    )
    suite_ending_task = build_suite_ending_task(suite, suite_ending_dependencies)

    ###
    # Return tasks != None
    ###
    task_iter = itertools.chain(
        (suite_beginning_task,),
        (suite_setup_task,), test_tasks, (suite_teardown_task,), sub_suite_tasks,
        (suite_ending_task,)
    )
    return list(filter(bool, task_iter))


class SuiteBeginningTask(BaseTask):
    def __init__(self, suite, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self._dependencies = dependencies

    def get_on_success_dependencies(self):
        return self._dependencies

    def run(self, context):
        context.session.start_suite(self.suite)

    def skip(self, context, _):
        self.run(context)

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_beginning_task(suite, dependencies):
    return SuiteBeginningTask(suite, dependencies)


class SuiteInitializationTask(BaseTask):
    def __init__(self, suite, setup_teardown_funcs, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self.setup_teardown_funcs = setup_teardown_funcs
        self._dependencies = dependencies
        self.teardown_funcs = []

    def get_on_success_dependencies(self):
        return self._dependencies

    def run(self, context):
        if any(setup for setup, _ in self.setup_teardown_funcs):
            # before actual initialization
            context.session.start_suite_setup(self.suite)
            context.session.set_step("Setup suite")
            # actual initialization
            self.teardown_funcs = context.run_setup_funcs(
                self.setup_teardown_funcs, ReportLocation.in_suite_setup(self.suite)
            )
            # after actual initialization
            context.session.end_suite_setup(self.suite)
            if not context.session.is_successful(ReportLocation.in_suite_setup(self.suite)):
                raise TaskFailure("suite '%s' setup failed" % self.suite.path)
        else:
            self.teardown_funcs = [teardown for _, teardown in self.setup_teardown_funcs if teardown]

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def wrap_setup_suite(suite, scheduled_fixtures):
    setup_suite = suite.get_hook("setup_suite")
    if setup_suite is None:
        return None

    fixtures_names = suite.get_hook_params("setup_suite")
    def wrapper():
        fixtures = scheduled_fixtures.get_fixture_results(fixtures_names)
        setup_suite(**fixtures)

    return wrapper


def build_suite_initialization_task(suite, scheduled_fixtures, dependencies, force_disabled):
    if not suite.has_enabled_tests() and not force_disabled:
        return None

    setup_teardown_funcs = []

    if not scheduled_fixtures.is_empty():
        setup_teardown_funcs.extend(scheduled_fixtures.get_setup_teardown_pairs())

    if suite.get_injected_fixture_names():
        setup_teardown_funcs.append((
            lambda: suite.inject_fixtures(scheduled_fixtures.get_fixture_results(suite.get_injected_fixture_names())),
            None
        ))

    if suite.has_hook("setup_suite") or suite.has_hook("teardown_suite"):
        setup_teardown_funcs.append([
            wrap_setup_suite(suite, scheduled_fixtures),
            suite.get_hook("teardown_suite")
        ])

    return SuiteInitializationTask(suite, setup_teardown_funcs, dependencies) if setup_teardown_funcs else None


class SuiteEndingTask(BaseTask):
    def __init__(self, suite, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self._dependencies = dependencies

    def get_on_success_dependencies(self):
        return self._dependencies

    def run(self, context):
        context.session.end_suite(self.suite)

    def skip(self, context, _):
        self.run(context)

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_ending_task(suite, dependencies):
    return SuiteEndingTask(suite, dependencies)


class SuiteTeardownTask(BaseTask):
    def __init__(self, suite, suite_setup_task, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self.suite_setup_task = suite_setup_task
        self._dependencies = dependencies

    def get_on_completion_dependencies(self):
        return self._dependencies

    def run(self, context):
        if any(self.suite_setup_task.teardown_funcs):
            # before actual teardown
            context.session.start_suite_teardown(self.suite)
            context.session.set_step("Teardown suite")
            # actual teardown
            context.run_teardown_funcs(self.suite_setup_task.teardown_funcs)
            # after actual teardown
            context.session.end_suite_teardown(self.suite)

    def skip(self, context, _):
        self.run(context)

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_teardown_task(suite, suite_setup_task, dependencies):
    return SuiteTeardownTask(suite, suite_setup_task, dependencies) if suite_setup_task else None


class TestSessionSetupTask(BaseTask):
    def __init__(self, scheduled_fixtures):
        BaseTask.__init__(self)
        self.scheduled_fixtures = scheduled_fixtures
        self.teardown_funcs = []

    def run(self, context):
        setup_teardown_funcs = self.scheduled_fixtures.get_setup_teardown_pairs()

        if any(setup for setup, _ in setup_teardown_funcs):
            # before actual setup
            context.session.start_test_session_setup()
            context.session.set_step("Setup test session")
            # actual setup
            self.teardown_funcs = context.run_setup_funcs(setup_teardown_funcs, ReportLocation.in_test_session_setup())
            # after actual setup
            context.session.end_test_session_setup()
            if not context.session.is_successful(ReportLocation.in_test_session_setup()):
                raise TaskFailure("test session setup failed")
        else:
            self.teardown_funcs = [teardown for _, teardown in setup_teardown_funcs if teardown]


def build_test_session_setup_task(scheduled_fixtures):
    return TestSessionSetupTask(scheduled_fixtures) if not scheduled_fixtures.is_empty() else None


class TestSessionTeardownTask(BaseTask):
    def __init__(self, test_session_setup_task, dependencies):
        BaseTask.__init__(self)
        self.test_session_setup_task = test_session_setup_task
        self._dependencies = dependencies

    def get_on_completion_dependencies(self):
        return self._dependencies

    def run(self, context):
        if any(self.test_session_setup_task.teardown_funcs):
            # before actual teardown
            context.session.start_test_session_teardown()
            context.session.set_step("Teardown test session")
            # actual teardown
            context.run_teardown_funcs(self.test_session_setup_task.teardown_funcs)
            # after actual teardown
            context.session.end_test_session_teardown()

    def skip(self, context, _):
        self.run(context)


def build_test_session_teardown_task(test_session_setup_task, dependencies):
    return TestSessionTeardownTask(test_session_setup_task, dependencies) if test_session_setup_task else None


def lookup_test_task(tasks, test_path):
    try:
        return next(task for task in tasks if isinstance(task, TestTask) and task.test.path == test_path)
    except StopIteration:
        raise LookupError("Cannot find test '%s' in tasks" % test_path)


def build_tasks(suites, fixture_registry, session_scheduled_fixtures, force_disabled):
    ###
    # Build test session setup task
    ###
    test_session_setup_task = build_test_session_setup_task(session_scheduled_fixtures)

    ###
    # Build suite tasks
    ###
    suite_tasks = []
    for suite in suites:
        suite_tasks.extend(
            build_suite_tasks(
                suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task,
                force_disabled=force_disabled
            )
        )

    ###
    # Build test session teardown task
    ###
    if test_session_setup_task:
        test_session_teardown_dependencies = [
            task for task in suite_tasks if isinstance(task, SuiteEndingTask) and task.suite in suites
        ]
        test_session_teardown_task = build_test_session_teardown_task(
            test_session_setup_task, test_session_teardown_dependencies
        )
    else:
        test_session_teardown_task = None

    ###
    # Get all effective tasks (task != None)
    ###
    task_iter = itertools.chain((test_session_setup_task,), suite_tasks, (test_session_teardown_task,))
    tasks = list(filter(bool, task_iter))

    ###
    # Add extra dependencies in tasks for tests that depend on other tests
    ###
    for test in flatten_tests(suites):
        if not test.dependencies:
            continue
        test_task = lookup_test_task(tasks, test.path)
        for dep_test_path in test.dependencies:
            try:
                dep_test = lookup_test_task(tasks, dep_test_path)
            except LookupError:
                raise UserError(
                    "Cannot find dependency test '%s' for '%s', "
                    "either the test does not exist or is not going to be run" % (dep_test_path, test.path)
                )
            test_task.dependencies.append(dep_test)

    ###
    # Return tasks
    ###
    return tasks


def _run_suites(suites, fixture_registry, pre_run_scheduled_fixtures, session,
                force_disabled=False, stop_on_failure=False, nb_threads=1):
    # build tasks and run context
    session_scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_session(
        suites, pre_run_scheduled_fixtures, force_disabled
    )
    tasks = build_tasks(suites, fixture_registry, session_scheduled_fixtures, force_disabled)
    context = RunContext(session, fixture_registry, force_disabled, stop_on_failure)

    with session.event_manager.handle_events():
        session.start_test_session()
        run_tasks(tasks, context, nb_threads)
        session.end_test_session()

    exception, serialized_exception = session.event_manager.get_pending_failure()
    if exception:
        raise exception.__class__(serialized_exception)


def run_suites(suites, fixture_registry, session, force_disabled=False, stop_on_failure=False, nb_threads=1):
    fixture_teardowns = []

    # setup of 'pre_run' fixtures
    errors = []
    scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_pre_run(suites, force_disabled)
    initialize_fixture_cache(scheduled_fixtures)
    for setup, teardown in scheduled_fixtures.get_setup_teardown_pairs():
        try:
            setup()
        except UserError:
            raise
        except Exception:
            errors.append("Got the following exception when executing fixture (scope 'pre_run')%s" % (
                serialize_current_exception(show_stacktrace=True)
            ))
            break
        fixture_teardowns.append(teardown)

    if not errors:
        _run_suites(
            suites, fixture_registry, scheduled_fixtures, session,
            force_disabled=force_disabled, stop_on_failure=stop_on_failure, nb_threads=nb_threads
        )

    # teardown of 'pre_run' fixtures
    for teardown in fixture_teardowns:
        try:
            teardown()
        except UserError:
            raise
        except Exception:
            errors.append("Got the following exception on fixture teardown (scope 'pre_run')%s" % (
                serialize_current_exception(show_stacktrace=True)
            ))

    if errors:
        raise LemoncheesecakeException("\n".join(errors))
    else:
        return session.report.is_successful()



'''
Created on Mar 27, 2017

@author: nicolas
'''

from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.composites import is_, not_
from lemoncheesecake.matching.matchers.types_ import is_bool


class EqualTo(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation("to be equal to %s" % jsonify(self.expected))

    def matches(self, actual):
        from lemoncheesecake.matching import DISPLAY_DETAILS_WHEN_EQUAL

        if actual == self.expected:
            if DISPLAY_DETAILS_WHEN_EQUAL:
                return MatchResult.success("got %s" % jsonify(actual))
            else:
                return MatchResult.success()
        else:
            return MatchResult.failure("got %s" % jsonify(actual))


def equal_to(expected):
    """Test if value is equal to expected"""
    return EqualTo(expected)


def _comparator(comparison_description, comparison_func):
    def wrapper(expected):
        class _Comparator(Matcher):
            def __init__(self, expected):
                self.expected = expected

            def build_description(self, transformation):
                return transformation("to be %s %s" % (comparison_description, jsonify(self.expected)))

            def matches(self, actual):
                return MatchResult(comparison_func(actual, self.expected), "got %s" % jsonify(actual))

        return _Comparator(expected)

    wrapper.__doc__ = """Test if value is %s expected""" % comparison_description
    return wrapper


not_equal_to = _comparator("not equal to", lambda a, e: a != e)

greater_than = _comparator("greater than", lambda a, e: a > e)
greater_than_or_equal_to = _comparator("greater than or equal to", lambda a, e: a >= e)

less_than = _comparator("less than", lambda a, e: a < e)
less_than_or_equal_to = _comparator("greater than or equal to", lambda a, e: a <= e)


class IsBetween(Matcher):
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def build_description(self, transformation):
        return transformation("to be between %s and %s" % (self.min, self.max))

    def matches(self, actual):
        return MatchResult(self.min <= actual <= self.max, "got %s" % jsonify(actual))


def is_between(min, max):
    # type: (float, float) -> IsBetween
    """Test if value is between min and max"""
    return IsBetween(min, max)


class IsNone(Matcher):
    def build_description(self, transformation):
        return transformation("to be null")

    def matches(self, actual):
        if actual is None:
            return MatchResult.success("got null")
        else:
            return MatchResult.failure("got %s" % jsonify(actual))


def is_none():
    """Test if value is None"""
    return IsNone()


def is_not_none():
    """Test if value is not None"""
    return not_(is_none())


class HasLength(Matcher):
    def __init__(self, matcher):
        # type: (Matcher) -> None
        self.matcher = matcher

    def build_description(self, transformation):
        return transformation(
            "to have a length that %s" % (self.matcher.build_description(MatcherDescriptionTransformer(conjugate=True)))
        )

    def matches(self, actual):
        return self.matcher.matches(len(actual))


def has_length(length):
    # type: (int) -> HasLength
    """Test if value has a length of"""
    return HasLength(is_(length))


def is_true():
    """Test if value is true (boolean type)"""
    return is_bool(True)


def is_false():
    """Test if value is false (boolean type)"""
    return is_bool(False)



from __future__ import print_function

import os
import sys

try:
    import slacker
    SLACKER_IS_AVAILABLE = True
except ImportError:
    SLACKER_IS_AVAILABLE = False

from lemoncheesecake.reporting import ReportingBackend, ReportingSession, ReportingSessionBuilderMixin, \
    check_report_message_template
from lemoncheesecake.exceptions import UserError


class SlackReportingSession(ReportingSession):
    def __init__(self, auth_token, channel, message_template, proxy=None, only_notify_failure=False):
        self.slacker = slacker.Slacker(auth_token, http_proxy=proxy, https_proxy=proxy)
        self.channel = channel
        self.message_template = message_template
        self.only_notify_failure = only_notify_failure

    def on_test_session_end(self, event):
        if self.only_notify_failure and event.report.is_successful():
            return

        message = event.report.build_message(self.message_template)
        try:
            self.slacker.chat.post_message(self.channel, message)
        except slacker.Error as excp:
            print("Error while notifying Slack channel/user '%s', got: %s" % (
                self.channel, excp
            ), file=sys.stderr)


def get_env_var(name, optional=False, default=None):
    try:
        return os.environ[name]
    except KeyError as excp:
        if optional:
            return default
        else:
            raise UserError("Slack reporting backend: missing environment variable '%s'" % excp)


class SlackReportingBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def get_name(self):
        return "slack"

    def is_available(self):
        return SLACKER_IS_AVAILABLE

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        try:
            message_template = check_report_message_template(get_env_var(
                "LCC_SLACK_MESSAGE_TEMPLATE", optional=True,
                default="{passed} passed, {failed} failed, {skipped} skipped"
            ))
        except ValueError as excp:
            raise UserError("Slack reporting backend: %s" % excp)

        return SlackReportingSession(
            auth_token=get_env_var("LCC_SLACK_AUTH_TOKEN"),
            channel=get_env_var("LCC_SLACK_CHANNEL"),
            message_template=message_template,
            proxy=get_env_var("LCC_SLACK_HTTP_PROXY", optional=True, default=None),
            only_notify_failure="LCC_SLACK_ONLY_NOTIFY_FAILURE" in os.environ
        )



'''
Created on Apr 4, 2017

@author: nicolas
'''

import re
import difflib
import json

from typing import Any, Pattern, Union, Callable

from lemoncheesecake.helpers.text import jsonify, STRING_TYPES
from lemoncheesecake.matching.matcher import Matcher, MatchResult

_REGEXP_TYPE = type(re.compile(r"dummy"))


def _is_string(actual):
    return type(actual) in STRING_TYPES


class StartsWith(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation('to start with "%s"' % self.expected)

    def matches(self, actual):
        return MatchResult(
            _is_string(actual) and actual.startswith(self.expected),
            "got %s" % jsonify(actual)
        )


def starts_with(expected):
    # type: (str) -> StartsWith
    """Test if string begins with given prefix"""
    return StartsWith(expected)


class EndsWith(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation('to end with "%s"' % self.expected)

    def matches(self, actual):
        return MatchResult(
            _is_string(actual) and actual.endswith(self.expected),
            "got %s" % jsonify(actual)
        )


def ends_with(expected):
    # type: (str) -> EndsWith
    """Test if string ends with given suffix"""
    return EndsWith(expected)


class ContainsString(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation('to contain "%s"' % self.expected)

    def matches(self, actual):
        return MatchResult(
            _is_string(actual) and self.expected in actual,
            "got %s" % jsonify(actual)
        )


def contains_string(expected):
    # type: (str) -> ContainsString
    """Test if string contains sub string"""
    return ContainsString(expected)


class MatchPattern(Matcher):
    def __init__(self, expected, description=None, mention_regexp=False):
        assert not (mention_regexp and description is None)
        self.expected = expected
        self.description = description
        self.mention_regexp = mention_regexp

    def build_description(self, transformation):
        if self.description:
            desc = transformation("to be %s" % self.description)
            if self.mention_regexp:
                desc += ' (pattern: "%s")' % self.expected.pattern
            return desc
        else:
            return transformation('to match pattern "%s"' % self.expected.pattern)

    def matches(self, actual):
        try:
            match = self.expected.search(actual)
        except TypeError:
            return MatchResult.failure("Invalid value %s (%s)" % (repr(actual), type(actual)))
        return MatchResult(match is not None, "got %s" % jsonify(actual))


def match_pattern(pattern, description=None, mention_regexp=False):
    # type: (Union[str, Pattern], str, bool) -> MatchPattern
    """Test if string matches given pattern (using the `search` method of the `re` module)"""
    return MatchPattern(
        pattern if type(pattern) == _REGEXP_TYPE else re.compile(pattern), description, mention_regexp
    )


def make_pattern_matcher(pattern, description=None, mention_regexp=False):
    # type: (Union[str, Pattern], str, bool) -> Callable[[], MatchPattern]
    def matcher():
        return match_pattern(pattern, description, mention_regexp)
    return matcher


def _diff_text(text1, text2, linesep):
    return "\n".join(
        difflib.unified_diff(
            text1.split(linesep), text2.split(linesep), fromfile="expected", tofile="actual", lineterm=""
        )
    )


class IsText(Matcher):
    def __init__(self, expected, linesep):
        self.expected = expected
        self.linesep = linesep

    def build_description(self, transformation):
        return transformation("to be the text:\n<<<\n%s\n>>>\n" % self.expected)

    def matches(self, actual):
        if actual == self.expected:
            return MatchResult.success()
        else:
            return MatchResult.failure(
                "Text does not match:\n%s" % _diff_text(self.expected, actual, self.linesep)
            )


def is_text(expected, linesep="\n"):
    # type: (str, str) -> IsText
    """
    Test if the two multi-lines texts match.

    If the two values do not match, the match result description will be a diff between the two texts.
    """
    return IsText(expected, linesep)


class IsJson(Matcher):
    def __init__(self, expected):
        self.expected = expected

    @staticmethod
    def _format_json(data):
        return json.dumps(data, ensure_ascii=False, sort_keys=True, indent=4)

    def build_description(self, transformation):
        return transformation("to be the JSON:\n%s\n" % self._format_json(self.expected))

    def matches(self, actual):
        if actual == self.expected:
            return MatchResult.success()
        else:
            return MatchResult.failure(
                "JSON does not match:\n%s" % _diff_text(
                    self._format_json(self.expected), self._format_json(actual), "\n"
                )
            )


def is_json(expected):
    # type: (Any) -> IsJson
    """
    Test if the two data structures (that can be represented as JSON) match.

    If the two values do not match, the match result description will be a textual diff between the two JSON
    representations.
    """
    return IsJson(expected)



'''
Created on Mar 27, 2016

@author: nicolas
'''

import re
import json
from collections import OrderedDict
import time

import lemoncheesecake
from lemoncheesecake.reporting.backend import FileReportBackend, ReportUnserializerMixin
from lemoncheesecake.reporting.report import (
    Report, Log, Check, Attachment, Url, Step, Result, TestResult, SuiteResult,
    format_time_as_iso8601, parse_iso8601_time
)
from lemoncheesecake.exceptions import ReportLoadingError

JS_PREFIX = "var reporting_data = "


def _serialize_time(t):
    return format_time_as_iso8601(t) if t is not None else None


def _odict(*args):
    d = OrderedDict()
    i = 0
    while i < len(args):
        d[args[i]] = args[i+1]
        i += 2
    return d


def _serialize_steps(steps):
    json_steps = []
    for step in steps:
        json_step = _odict(
            "description", step.description,
            "start_time", _serialize_time(step.start_time),
            "end_time", _serialize_time(step.end_time),
            "entries", []
        )
        json_steps.append(json_step)
        for log in step.get_logs():
            if isinstance(log, Log):
                json_log = _odict(
                    "type", "log",
                    "level", log.level,
                    "message", log.message,
                    "time", _serialize_time(log.time)
                )
            elif isinstance(log, Attachment):
                json_log = _odict(
                    "type", "attachment",
                    "description", log.description,
                    "filename", log.filename,
                    "as_image", log.as_image,
                    "time", _serialize_time(log.time)
                )
            elif isinstance(log, Url):
                json_log = _odict(
                    "type", "url",
                    "description", log.description,
                    "url", log.url,
                    "time", _serialize_time(log.time)
                )
            elif isinstance(log, Check):
                json_log = _odict(
                    "type", "check",
                    "description", log.description,
                    "is_successful", log.is_successful,
                    "details", log.details,
                    "time", _serialize_time(log.time)
                )
            else:
                raise ValueError("Don't know how to handle step log %s" % log)
            json_step["entries"].append(json_log)
    return json_steps


def _serialize_node_metadata(node, json_node):
    json_node.update(_odict(
        "name", node.name, "description", node.description,
        "tags", node.tags,
        "properties", node.properties,
        "links", [_odict("name", link[1], "url", link[0]) for link in node.links]
    ))


def _serialize_result(result):
    return _odict(
        "start_time", _serialize_time(result.start_time),
        "end_time", _serialize_time(result.end_time),
        "steps", _serialize_steps(result.get_steps()),
        "status", result.status,
        "status_details", result.status_details
    )


def _serialize_test_result(test):
    json_test = _serialize_result(test)
    _serialize_node_metadata(test, json_test)
    return json_test


def _serialize_suite_result(suite):
    json_suite = _odict(
        "start_time", _serialize_time(suite.start_time),
        "end_time", _serialize_time(suite.end_time),
        "tests", list(map(_serialize_test_result, suite.get_tests())),
        "suites", list(map(_serialize_suite_result, suite.get_suites()))
    )
    _serialize_node_metadata(suite, json_suite)
    if suite.suite_setup:
        json_suite["suite_setup"] = _serialize_result(suite.suite_setup)
    if suite.suite_teardown:
        json_suite["suite_teardown"] = _serialize_result(suite.suite_teardown)

    return json_suite


def serialize_report_into_json(report):
    json_report = _odict(
        "lemoncheesecake_version", lemoncheesecake.__version__,
        "report_version", 1.1,
        "start_time", _serialize_time(report.start_time),
        "end_time", _serialize_time(report.end_time),
        "generation_time", _serialize_time(time.time()),
        "nb_threads", report.nb_threads,
        "title", report.title,
        "info", [[n, v] for n, v in report.info]
    )

    if report.test_session_setup:
        json_report["test_session_setup"] = _serialize_result(report.test_session_setup)

    json_report["suites"] = list(map(_serialize_suite_result, report.get_suites()))

    if report.test_session_teardown:
        json_report["test_session_teardown"] = _serialize_result(report.test_session_teardown)

    return json_report


def save_report_into_file(report, filename, javascript_compatibility=True, pretty_formatting=False):
    json_report = serialize_report_into_json(report)
    with open(filename, "w") as fh:
        if javascript_compatibility:
            fh.write(JS_PREFIX)
        if pretty_formatting:
            fh.write(json.dumps(json_report, indent=4))
        else:
            fh.write(json.dumps(json_report))


def _unserialize_time(t):
    return parse_iso8601_time(t) if t is not None else None


def _unserialize_step(json_step):
    step = Step(json_step["description"])
    step.start_time = _unserialize_time(json_step["start_time"])
    step.end_time = _unserialize_time(json_step["end_time"])
    for json_log in json_step["entries"]:
        if json_log["type"] == "log":
            step_log = Log(
                json_log["level"], json_log["message"], _unserialize_time(json_log["time"])
            )
        elif json_log["type"] == "attachment":
            step_log = Attachment(
                json_log["description"], json_log["filename"], json_log["as_image"],
                _unserialize_time(json_log["time"])
            )
        elif json_log["type"] == "url":
            step_log = Url(
                json_log["description"], json_log["url"], _unserialize_time(json_log["time"])
            )
        elif json_log["type"] == "check":
            step_log = Check(
                json_log["description"], json_log["is_successful"], json_log["details"],
                _unserialize_time(json_log["time"])
            )
        else:
            raise ValueError("Unknown step log type '%s'" % json_log["type"])
        step.add_log(step_log)
    return step


def _unserialize_result(json_result, result):
    result.status = json_result["status"]
    # status_details for non-test results has been introduced in report version 1.1:
    result.status_details = json_result.get("status_details", None)
    result.start_time = _unserialize_time(json_result["start_time"])
    result.end_time = _unserialize_time(json_result["end_time"])
    for json_step in json_result["steps"]:
        result.add_step(_unserialize_step(json_step))

    return result


def _unserialize_node_metadata(json_node, node):
    node.tags = json_node["tags"]
    node.properties = json_node["properties"]
    node.links = [(link["url"], link["name"]) for link in json_node["links"]]


def _unserialize_test_result(json_test):
    test = TestResult(json_test["name"], json_test["description"])
    _unserialize_result(json_test, test)
    _unserialize_node_metadata(json_test, test)
    return test


def _unserialize_suite_result(json_suite):
    suite = SuiteResult(json_suite["name"], json_suite["description"])
    _unserialize_node_metadata(json_suite, suite)
    suite.start_time = _unserialize_time(json_suite["start_time"])
    suite.end_time = _unserialize_time(json_suite["end_time"])

    if "suite_setup" in json_suite:
        suite.suite_setup = Result()
        _unserialize_result(json_suite["suite_setup"], suite.suite_setup)

    for json_test in json_suite["tests"]:
        suite.add_test(_unserialize_test_result(json_test))

    if "suite_teardown" in json_suite:
        suite.suite_teardown = Result()
        _unserialize_result(json_suite["suite_teardown"], suite.suite_teardown)

    for json_sub_suite in json_suite["suites"]:
        suite.add_suite(_unserialize_suite_result(json_sub_suite))

    return suite


def _unserialize_report(json_report):
    report = Report()

    report.title = json_report["title"]
    report.info = json_report["info"]
    report.start_time = _unserialize_time(json_report["start_time"])
    report.end_time = _unserialize_time(json_report["end_time"])
    report.saving_time = _unserialize_time(json_report["generation_time"])
    report.nb_threads = json_report["nb_threads"]

    if "test_session_setup" in json_report:
        report.test_session_setup = Result()
        _unserialize_result(json_report["test_session_setup"], report.test_session_setup)

    for json_suite in json_report["suites"]:
        report.add_suite(_unserialize_suite_result(json_suite))

    if "test_session_teardown" in json_report:
        report.test_session_teardown = Result()
        _unserialize_result(json_report["test_session_teardown"], report.test_session_teardown)

    return report


def load_report_from_file(filename):
    try:
        with open(filename, "r") as fh:
            js_content = fh.read()
    except IOError as e:
        raise e  # re-raise as-is

    js_content = re.sub("^" + JS_PREFIX, "", js_content)

    try:
        js = json.loads(js_content)
    except ValueError as e:
        raise ReportLoadingError(str(e))

    report_version = js.get("report_version")
    if report_version is None:
        raise ReportLoadingError("Cannot find 'report_version' in JSON")
    if report_version >= 2.0:
        raise ReportLoadingError("Incompatible report version: got %s while 1.x is supported" % report_version)

    return _unserialize_report(js)


class JsonBackend(FileReportBackend, ReportUnserializerMixin):
    def __init__(self, javascript_compatibility=True, pretty_formatting=False):
        self.javascript_compatibility = javascript_compatibility
        self.pretty_formatting = pretty_formatting

    def get_name(self):
        return "json"

    def get_report_filename(self):
        return "report.js"

    def save_report(self, filename, report):
        save_report_into_file(
            report, filename,
            javascript_compatibility=self.javascript_compatibility, pretty_formatting=self.pretty_formatting
        )

    def load_report(self, path):
        report = load_report_from_file(path)
        report.bind(self, path)
        return report



import os
import os.path as osp
from shutil import copy, copytree

from lemoncheesecake.helpers.resources import get_resource_path
from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSessionBuilderMixin, ReportingSession
from lemoncheesecake.helpers.console import bold


class HtmlReportWriter(ReportingSession):
    def __init__(self, report_dir, fat_html):
        self.report_dir = report_dir
        self.fat_html = fat_html

    def on_test_session_start(self, _):
        src_dir = get_resource_path("html")
        dst_dir = osp.join(self.report_dir, ".html")

        os.mkdir(dst_dir)
        copy(osp.join(src_dir, "report.js"), dst_dir)
        copy(osp.join(src_dir, "report.css"), dst_dir)
        copy(osp.join(src_dir, "logo.png"), dst_dir)

        if self.fat_html:
            copy(osp.join(src_dir, "report_static.html"), osp.join(self.report_dir, "report.html"))
            copy(osp.join(src_dir, "jquery-1.12.3.min.js"), dst_dir)
            copytree(osp.join(src_dir, "bootstrap-3.3.6-dist"), osp.join(dst_dir, "bootstrap-3.3.6-dist"))
            copy(osp.join(src_dir, "bootstrap-slate.min.css"), osp.join(dst_dir, "bootstrap-3.3.6-dist", "css", "bootstrap.min.css"))
        else:
            copy(osp.join(src_dir, "report_external.html"), osp.join(self.report_dir, "report.html"))

    def on_test_session_end(self, _):
        print("%s : file://%s/report.html" % (bold("HTML report"), self.report_dir))
        print()


class HtmlBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def __init__(self):
        self.fat_html = True

    def get_name(self):
        return "html"

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        return HtmlReportWriter(report_dir, self.fat_html)



'''
Created on Sep 8, 2016

@author: nicolas
'''

from lemoncheesecake.exceptions import SuiteLoadingError
from lemoncheesecake.helpers.orderedset import OrderedSet
from lemoncheesecake.helpers.introspection import get_callable_args, get_object_attributes
from lemoncheesecake.testtree import BaseTest, BaseSuite

SUITE_HOOKS = "setup_test", "teardown_test", "setup_suite", "teardown_suite"


def _is_node_disabled(node):
    while node is not None:
        if node.disabled:
            return node.disabled
        node = node.parent_suite
    return False


class Test(BaseTest):
    """
    Internal representation of a test.
    """
    def __init__(self, name, description, callback):
        BaseTest.__init__(self, name, description)
        self.callback = callback
        self.disabled = False
        self.hidden = False
        self.rank = 0
        self.dependencies = []
        self.parameters = {}

    def is_disabled(self):
        return _is_node_disabled(self)

    def is_enabled(self):
        return not self.is_disabled()

    def get_arguments(self):
        return get_callable_args(self.callback)

    def get_fixtures(self):
        return list(filter(lambda arg: arg not in self.parameters, self.get_arguments()))


class InjectedFixture:
    def __init__(self, fixture_name):
        self.fixture_name = fixture_name


class Suite(BaseSuite):
    """
    Internal representation of a suite.
    """
    def __init__(self, obj, name, description):
        BaseSuite.__init__(self, name, description)
        self.obj = obj
        self.rank = 0
        self.disabled = False
        self.hidden = False
        self._hooks = {}
        self._injected_fixtures = self._load_injected_fixtures(obj) if obj else {}
        # to optimize unique constraint checks on test/suite name/description, keep those
        # strings in sets:
        self._test_names = set()
        self._test_descriptions = set()
        self._sub_suite_names = set()
        self._sub_suite_descriptions = set()

    @staticmethod
    def _assert_hook_name(hook_name):
        assert hook_name in SUITE_HOOKS, "Invalid hook name '%s'" % hook_name

    @staticmethod
    def _load_injected_fixtures(obj):
        fixtures = {}
        for attr_name, attr in get_object_attributes(obj):
            if isinstance(attr, InjectedFixture):
                fixtures[attr.fixture_name or attr_name] = attr_name
        return fixtures

    def is_disabled(self):
        return _is_node_disabled(self)

    def has_enabled_tests(self):
        return any(test.is_enabled() for test in self.get_tests())

    def add_hook(self, hook_name, func):
        self._assert_hook_name(hook_name)
        self._hooks[hook_name] = func

    def has_hook(self, hook_name):
        self._assert_hook_name(hook_name)
        return hook_name in self._hooks

    def get_hook(self, hook_name):
        self._assert_hook_name(hook_name)
        return self._hooks.get(hook_name)

    def get_hook_params(self, hook_name):
        hook = self.get_hook(hook_name)
        assert hook
        return get_callable_args(hook)

    def get_injected_fixture_names(self):
        return self._injected_fixtures.keys()

    def inject_fixtures(self, fixtures):
        for fixture_name, fixture_value in fixtures.items():
            attr_name = self._injected_fixtures[fixture_name]
            setattr(self.obj, attr_name, fixture_value)

    def add_test(self, test):
        if test.description in self._test_descriptions:
            raise SuiteLoadingError(
                "A test with description '%s' is already registered in test suite %s" % (test.description, self.path)
            )
        if test.name in self._test_names:
            raise SuiteLoadingError(
                "A test with name '%s' is already registered in test suite %s" % (test.name, self.path)
            )

        BaseSuite.add_test(self, test)
        self._test_names.add(test.name)
        self._test_descriptions.add(test.description)

    def add_suite(self, suite):
        if suite.description in self._sub_suite_descriptions:
            raise SuiteLoadingError(
                "A sub test suite with description '%s' is already registered in test suite %s" % (suite.name, self.path)
            )
        if suite.name in self._sub_suite_names:
            raise SuiteLoadingError(
                "A sub test suite with name '%s' is already registered in test suite %s" % (suite.name, self.path)
            )

        BaseSuite.add_suite(self, suite)
        self._sub_suite_names.add(suite.name)
        self._sub_suite_descriptions.add(suite.description)

    def pull_node(self):
        suite = BaseSuite.pull_node(self)
        suite._test_names = set()
        suite._test_descriptions = set()
        suite._sub_suite_names = set()
        suite._sub_suite_descriptions = set()
        return suite

    def get_fixtures(self):
        fixtures = OrderedSet(self._injected_fixtures.keys())

        suite_setup = self.get_hook("setup_suite")
        if suite_setup:
            fixtures.update(get_callable_args(suite_setup))

        return fixtures



'''
Created on Jan 24, 2016

@author: nicolas
'''

from lemoncheesecake.reporting.report import \
    SuiteResult, TestResult, Result, Step, Log, Check, Attachment, Url


class ReportWriter:
    def __init__(self, report):
        self.report = report
        self.active_steps = {}

    def _get_test_result(self, test):
        return self.report.get_test(test)

    def _get_suite_result(self, suite):
        return self.report.get_suite(suite)

    def _add_step_log(self, log, event):
        result = self.report.get(event.location)
        assert result, "Cannot find location %s in the report" % event.location
        step = self._lookup_step(event)
        assert step, "Cannot find active step for %s" % event.location
        assert not step.end_time, "Cannot update step '%s', it has already been ended" % step.description
        step.add_log(log)

    @staticmethod
    def _initialize_result(start_time):
        result = Result()
        result.start_time = start_time
        return result

    @staticmethod
    def _initialize_test_result(test, start_time):
        result = TestResult(test.name, test.description)
        result.tags.extend(test.tags)
        result.properties.update(test.properties)
        result.links.extend(test.links)
        result.rank = test.rank
        result.start_time = start_time
        return result

    @staticmethod
    def _finalize_result(result, end_time):
        result.end_time = end_time
        result.status = "passed" if result.is_successful() else "failed"

    def _lookup_step(self, event):
        try:
            return self.active_steps[event.thread_id]
        except KeyError:
            return None

    def on_test_session_start(self, event):
        self.report.start_time = event.time

    def on_test_session_end(self, event):
        self.report.end_time = event.time

    def on_test_session_setup_start(self, event):
        self.report.test_session_setup = self._initialize_result(event.time)

    def on_test_session_setup_end(self, event):
        self._finalize_result(self.report.test_session_setup, event.time)

    def on_test_session_teardown_start(self, event):
        self.report.test_session_teardown = self._initialize_result(event.time)

    def on_test_session_teardown_end(self, event):
        self._finalize_result(self.report.test_session_teardown, event.time)

    def on_suite_start(self, event):
        suite = event.suite
        suite_result = SuiteResult(suite.name, suite.description)
        suite_result.start_time = event.time
        suite_result.tags.extend(suite.tags)
        suite_result.properties.update(suite.properties)
        suite_result.links.extend(suite.links)
        suite_result.rank = suite.rank
        if suite.parent_suite:
            parent_suite_result = self._get_suite_result(suite.parent_suite)
            parent_suite_result.add_suite(suite_result)
        else:
            self.report.add_suite(suite_result)

    def on_suite_end(self, event):
        suite_result = self._get_suite_result(event.suite)
        suite_result.end_time = event.time

    def on_suite_setup_start(self, event):
        suite_result = self._get_suite_result(event.suite)
        suite_result.suite_setup = self._initialize_result(event.time)

    def on_suite_setup_end(self, event):
        suite_result = self._get_suite_result(event.suite)
        self._finalize_result(suite_result.suite_setup, event.time)

    def on_suite_teardown_start(self, event):
        suite_result = self._get_suite_result(event.suite)
        suite_result.suite_teardown = self._initialize_result(event.time)

    def on_suite_teardown_end(self, event):
        suite_result = self._get_suite_result(event.suite)
        self._finalize_result(suite_result.suite_teardown, event.time)

    def on_test_start(self, event):
        test_result = self._initialize_test_result(event.test, event.time)
        suite_result = self._get_suite_result(event.test.parent_suite)
        suite_result.add_test(test_result)

    def on_test_end(self, event):
        test_result = self._get_test_result(event.test)
        self._finalize_result(test_result, event.time)

    def _bypass_test(self, test, status, status_details, time):
        test_result = self._initialize_test_result(test, time)
        test_result.end_time = time
        test_result.status = status
        test_result.status_details = status_details

        suite_result = self._get_suite_result(test.parent_suite)
        suite_result.add_test(test_result)

    def on_test_skipped(self, event):
        self._bypass_test(event.test, "skipped", event.skipped_reason, event.time)

    def on_test_disabled(self, event):
        self._bypass_test(event.test, "disabled", event.disabled_reason, event.time)

    def on_step_start(self, event):
        result = self.report.get(event.location)
        step = Step(event.step_description)
        step.start_time = event.time
        result.add_step(step)
        self.active_steps[event.thread_id] = step

    def on_step_end(self, event):
        step = self._lookup_step(event)
        step.end_time = event.time

    def on_log(self, event):
        self._add_step_log(
            Log(event.log_level, event.log_message, event.time), event
        )

    def on_check(self, event):
        self._add_step_log(
            Check(event.check_description, event.check_is_successful, event.check_details, event.time), event
        )

    def on_log_attachment(self, event):
        self._add_step_log(
            Attachment(event.attachment_description, event.attachment_path, event.as_image, event.time), event
        )

    def on_log_url(self, event):
        self._add_step_log(
            Url(event.url_description, event.url, event.time), event
        )



"""
This module gather public API to be used within suites, this module can be imported
as this for convenience:
import lemoncheesecake.api as lcc
"""

from lemoncheesecake.suite import Test, add_test_into_suite, \
    get_metadata, suite, test, tags, prop, link, disabled, visible_if, hidden, depends_on, inject_fixture, parametrized
from lemoncheesecake.session import set_step, detached_step, end_step, log_debug, log_info, log_warning, log_error, \
    log_check, prepare_attachment, prepare_image_attachment, save_attachment_file, save_image_file, \
    save_attachment_content, save_image_content, log_url, add_report_info, Thread
from lemoncheesecake.helpers.threading import ThreadedFactory
from lemoncheesecake.fixture import fixture, get_fixture
from lemoncheesecake.exceptions import AbortTest, AbortSuite, AbortAllTests, UserError

# for pydoc & sphinx
__all__ = [sym_name for sym_name in dir() if not sym_name.startswith("_")]



'''
Created on Sep 8, 2016

@author: nicolas
'''

import inspect
import copy

from typing import Any, Iterable, Callable, Optional, Tuple, Sequence, Union

from lemoncheesecake.helpers.text import STRING_TYPES
from lemoncheesecake.suite.core import InjectedFixture, Test


class Metadata(object):
    _next_rank = 1

    def __init__(self):
        self.is_test = False
        self.is_suite = False
        self.name = None
        self.description = None
        self.properties = {}
        self.tags = []
        self.links = []
        self.rank = 0
        self.dependencies = []
        self.disabled = False
        self.condition = None
        self.parametrized = None


def _get_metadata_next_rank():
    rank = Metadata._next_rank
    Metadata._next_rank += 1
    return rank


def build_description_from_name(name):
    return name.capitalize().replace("_", " ")


def add_test_into_suite(test, suite):
    # type: (Test, Any) -> None
    """
    Add test into suite

    :param test: a :py:class:`Test <lemoncheesecake.suite.core.Test>` instance
    :param suite: a suite decorated class instance (in that case the function must be
        called when the class is instantiated) or a module marked as a suite (in that case the function
        must be called when the module is loaded)
    """
    if not hasattr(suite, "_lccgeneratedtests"):
        suite._lccgeneratedtests = []
    if test.rank is None:
        test.rank = _get_metadata_next_rank()
    suite._lccgeneratedtests.append(test)


_objects_with_metadata = []
def get_metadata(obj):
    global _objects_with_metadata

    if hasattr(obj, "_lccmetadata"):
        if obj not in _objects_with_metadata:  # metadata comes from the superclass
            obj._lccmetadata = copy.deepcopy(obj._lccmetadata)
        return obj._lccmetadata
    else:
        obj._lccmetadata = Metadata()
        _objects_with_metadata.append(obj)
        return obj._lccmetadata


def suite(description=None, name=None, rank=None):
    """
    Decorator, mark a class as a suite class.

    :param description: suite's description (by default, the suite's description is built from the name)
    :param name: suite's name (by default, the suite's name is taken from the class's name)
    :param rank: this value is used to order suites of the same hierarchy level
    """
    def wrapper(klass):
        assert inspect.isclass(klass), "%s is not a class (suite decorator can only be used on a class)" % klass
        md = get_metadata(klass)
        assert not md.dependencies, "'depends_on' can not be used on a suite class"
        md.is_suite = True
        md.rank = rank if rank is not None else _get_metadata_next_rank()
        md.name = name or klass.__name__
        md.description = description or build_description_from_name(md.name)
        return klass
    return wrapper


def test(description=None, name=None):
    """
    Decorator, mark a function/method as a test.

    :param description: test's description
    :param name: test's name (by default, the suite's name is taken from the class's name)
    """
    def wrapper(func):
        assert inspect.isfunction(func), "%s is not a function (test decorator can only be used on a function)" % func
        md = get_metadata(func)
        md.is_test = True
        md.rank = _get_metadata_next_rank()
        md.name = name or func.__name__
        md.description = description or build_description_from_name(md.name)
        return func
    return wrapper


def tags(*tag_names):
    """Decorator, add tags to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.tags.extend(tag_names)
        return obj
    return wrapper


def prop(key, value):
    """Decorator, add a property (key/value) to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.properties[key] = value
        return obj
    return wrapper


def link(url, name=None):
    """Decorator, set a link (with an optional friendly name) to a test or a suite."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.links.append((url, name))
        return obj
    return wrapper


def disabled(reason=None):
    """Decorator, mark a test or a suite as disabled, meaning it won't be executed but will be visible
    in report. An optional reason can be passed to the decorator (*new in version 1.1.0*)."""
    def wrapper(obj):
        md = get_metadata(obj)
        md.disabled = reason if reason else True
        return obj
    return wrapper


def visible_if(condition):
    """
    Decorator, the test or suite will only appear if the given callable return a true value.

    :param condition: a callable that will take the test object if applied to a test or the suite class
        instance if applied to a suite.
    """
    def wrapper(obj):
        md = get_metadata(obj)
        md.condition = condition
        return obj
    return wrapper


def hidden():
    """Decorator, the test or suite won't be visible in the resulting test tree."""
    return visible_if(lambda _: False)


def depends_on(*deps):
    # type: (*str) -> Any
    """
    Decorator, only applicable to a test. Add dependencies to a test.

    :param deps: the test paths that the decorated test is depending on.
    """
    def wrapper(obj):
        md = get_metadata(obj)
        assert not md.is_suite, "'depends_on' can only be used on a test, not on a suite"
        md.dependencies.extend(deps)
        return obj
    return wrapper


def inject_fixture(fixture_name=None):
    """
    Inject a fixture into a suite. If no fixture name is specified then the name of the variable holding
    the injected fixture will be used.
    """
    return InjectedFixture(fixture_name)


def _default_naming_scheme(name, description, parameters, nb):
    return name + "_%d" % nb, description + " #%d" % nb


def _format_naming_scheme(name_fmt, description_fmt):
    def naming_scheme(_, __, parameters, ___):
        return name_fmt.format(**parameters), description_fmt.format(**parameters)
    return naming_scheme


class _Parametrized(object):
    def __init__(self, parameters_source, naming_scheme):
        self._parameters_source = parameters_source
        self.naming_scheme = naming_scheme

    @property
    def parameters_source(self):
        source = iter(self._parameters_source)
        try:
            first_item = next(source)
        except StopIteration:
            return
        if type(first_item) is dict:
            yield first_item
            for item in source:
                yield item
        else:
            if type(first_item) in STRING_TYPES:
                names = [s.strip() for s in first_item.split(",")]
            else:
                names = first_item  # assume list or tuple
            for values in source:
                yield dict(zip(names, values))


def parametrized(parameter_source, naming_scheme=_default_naming_scheme):
    # type: (Iterable, Optional[Union[Callable[[str, str, dict, int], Tuple[str, str]], Sequence]]) -> Any
    """
    Decorator, make the test parametrized.

    :param parameter_source: it can be either:

        -  an iterable of dicts, each dict representing the arguments passed to the test

        - a CSV-like mode, where the first element of the list represents the argument names as an str or a sequence
          (example: ``"i,j"`` or ``("i", "j")``) and the remaining elements are sequences of the arguments to be
          passed to the test.

    :param naming_scheme: optional, it can be either:

      - a optional function that takes as parameters the base test name, description, parameters, index
        and must return the expanded test name and description as a two elements list

      - a tuple/list of two (name + description) format strings, example: ``("test_{i}_plus_{j}", "Test {i} plus {j}")``
    """

    def wrapper(obj):
        md = get_metadata(obj)
        md.parametrized = _Parametrized(
            parameter_source,
            naming_scheme if callable(naming_scheme) else _format_naming_scheme(*naming_scheme)
        )
        return obj
    return wrapper



import six


def check_type(value_description, value, type_description, check, optional=False, show_actual_type=True):
    if optional and value is None:
        return

    if not check(value):
        error = "expect %s for %s" % (type_description, value_description)
        if show_actual_type:
            error += ", got %s" % value.__class__.__name__
        raise TypeError(error)


def check_type_string(value_description, value, optional=False):
    return check_type(
        value_description, value, "string", lambda value: isinstance(value, six.string_types), optional=optional
    )


def check_type_bool(value_description, value):
    return check_type(
        value_description, value, "boolean", lambda value: type(value) is bool
    )


def check_type_dict(value_description, value):
    return check_type(
        value_description, value, "dict", lambda value: isinstance(value, dict)
    )



import os
import os.path as osp
from typing import Sequence, Iterator

from lemoncheesecake.exceptions import ReportLoadingError
from lemoncheesecake.reporting.report import Report
from lemoncheesecake.reporting.backend import get_reporting_backends, ReportUnserializerMixin, ReportingBackend


def load_report_from_file(filename, backends=None):
    # type: (str, Sequence[ReportingBackend]) -> Report
    if backends is None:
        backends = get_reporting_backends()
    for backend in backends:
        if isinstance(backend, ReportUnserializerMixin):
            try:
                return backend.load_report(filename)
            except IOError as excp:
                raise ReportLoadingError("Cannot load report from file '%s': %s" % (filename, excp))
            except ReportLoadingError:
                pass
    raise ReportLoadingError("Cannot find any suitable report backend to load report file '%s'" % filename)


def load_reports_from_dir(dirname, backends=None):
    # type: (str, Sequence[ReportingBackend]) -> Iterator[Report]
    for filename in [os.path.join(dirname, filename) for filename in os.listdir(dirname)]:
        if os.path.isfile(filename):
            try:
                yield load_report_from_file(filename, backends)
            except ReportLoadingError:
                pass


def load_report(path, backends=None):
    # type: (str, Sequence[ReportingBackend]) -> Report
    """
    Load report from a report directory or file.
    """
    if osp.isdir(path):
        try:
            return next(load_reports_from_dir(path, backends))
        except StopIteration:
            raise ReportLoadingError("Cannot find any report in directory '%s'" % path)
    else:
        return load_report_from_file(path, backends)



'''
Created on Jun 16, 2017

@author: nicolas
'''

import copy
from collections import OrderedDict

from typing import Union, Tuple, List, Sequence, TypeVar, Generator

from lemoncheesecake.helpers.orderedset import OrderedSet


# Please note that attributes from base classes do not appear
# in generated api documentation
# (see https://github.com/sphinx-doc/sphinx/issues/741)

class BaseTreeNode(object):
    """
    :var name: name
    """

    def __init__(self, name, description):
        self.parent_suite = None
        #: name
        self.name = name
        #: description
        self.description = description
        #: tags, as a list
        self.tags = []
        #: properties, as a dict
        self.properties = {}
        #: links, as a list
        self.links = []

    @property
    def hierarchy(self):
        if self.parent_suite is not None:
            for node in self.parent_suite.hierarchy:
                yield node
        yield self

    @property
    def hierarchy_depth(self):
        # type: () -> int
        return len(list(self.hierarchy)) - 1

    @property
    def path(self):
        # type: () -> str
        """
        The complete path of the test node (example: if used on a test named "my_test" and a suite named
        "my_suite", then the path is "my_suite.my_test").
        """
        return ".".join(node.name for node in self.hierarchy)

    @property
    def hierarchy_paths(self):
        return (node.path for node in self.hierarchy)

    @property
    def hierarchy_descriptions(self):
        return (node.description for node in self.hierarchy)

    @property
    def hierarchy_tags(self):
        tags = OrderedSet()
        for node in self.hierarchy:
            tags.update(node.tags)
        return tags

    @property
    def hierarchy_properties(self):
        properties = {}
        for node in self.hierarchy:
            properties.update(node.properties)
        return properties

    @property
    def hierarchy_links(self):
        links = OrderedSet()
        for node in self.hierarchy:
            links.update(node.links)
        return links

    def pull_node(self):
        # type: () -> "_N"
        node = copy.copy(self)
        node.parent_suite = None
        return node

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.path)


_N = TypeVar("_N", bound=BaseTreeNode)


TreeNodeHierarchy = Union[Tuple[str, ...], List, BaseTreeNode, str]


def normalize_node_hierarchy(hierarchy):
    # type: (TreeNodeHierarchy) -> Tuple[str, ...]
    if type(hierarchy) is tuple:
        return hierarchy
    elif type(hierarchy) is list:
        return tuple(hierarchy)
    elif isinstance(hierarchy, BaseTreeNode):
        return tuple(p.name for p in hierarchy.hierarchy)
    else:  # assume str
        return tuple(hierarchy.split("."))


class BaseTest(BaseTreeNode):
    pass


T = TypeVar("T", bound=BaseTest)


class BaseSuite(BaseTreeNode):
    def __init__(self, name, description):
        BaseTreeNode.__init__(self, name, description)
        # NB: use OrderedDict instead of list to enable fast test lookup in suites
        # containing a large number of tests
        self._tests = OrderedDict()
        self._suites = []

    def add_test(self, test):
        """
        Add test to the suite.
        """
        test.parent_suite = self
        self._tests[test.name] = test

    def get_tests(self):
        """
        Get suite's tests.
        """
        return list(self._tests.values())

    def get_test_by_name(self, name):
        return self._tests[name]

    def add_suite(self, suite):
        """
        Add a sub-suite to the suite.
        """
        suite.parent_suite = self
        self._suites.append(suite)

    def get_suites(self):
        return self._suites

    def is_empty(self):
        if len(self.get_tests()) != 0:
            return False

        for sub_suite in self.get_suites():
            if not sub_suite.is_empty():
                return False

        return True

    def pull_node(self):
        # type: () -> "BaseSuite"
        node = BaseTreeNode.pull_node(self)
        node._tests = OrderedDict()
        node._suites = []
        return node

    def filter(self, test_filter):
        suite = self.pull_node()

        for test in self.get_tests():
            if test_filter(test):
                suite.add_test(test.pull_node())

        for sub_suite in filter_suites(self.get_suites(), test_filter):
            suite.add_suite(sub_suite)

        return suite


def filter_suites(suites, test_filter):
    return list(
        filter(
            lambda s: not s.is_empty(), (s.filter(test_filter) for s in suites)
        )
    )


S = TypeVar("S", bound=BaseSuite)


def flatten_suites(suites):
    # type: (Sequence[S]) -> Generator[S]
    for suite in suites:
        yield suite
        for sub_suite in flatten_suites(suite.get_suites()):
            yield sub_suite


def flatten_tests(suites):
    # type: (Sequence[S]) -> Generator[T]
    for suite in flatten_suites(suites):
        for test in suite.get_tests():
            yield test


def find_suite(suites, hierarchy):
    # type: (Sequence[S], TreeNodeHierarchy) -> S

    hierarchy = normalize_node_hierarchy(hierarchy)

    lookup_suites = suites
    lookup_suite = None
    for lookup_suite_name in hierarchy:
        try:
            lookup_suite = next(s for s in lookup_suites if s.name == lookup_suite_name)
        except StopIteration:
            raise LookupError("Cannot find suite named '%s' within %s" % (
                lookup_suite_name, [s.name for s in lookup_suites]
            ))

        lookup_suites = lookup_suite.get_suites()

    return lookup_suite


def find_test(suites, hierarchy):
    # type: (Sequence[BaseSuite], TreeNodeHierarchy) -> T

    hierarchy = normalize_node_hierarchy(hierarchy)

    suite = find_suite(suites, hierarchy[:-1])
    try:
        return suite.get_test_by_name(hierarchy[-1])
    except KeyError:
        raise LookupError("Cannot find test named '%s'" % ".".join(hierarchy))



from __future__ import print_function

import threading

from typing import Iterable

from lemoncheesecake.reporting import Report, Step, TestResult, SuiteResult, Log, Attachment, Url, Check, ReportLocation
from lemoncheesecake import events
from lemoncheesecake.events import EventManager


def _replay_step(location, step, eventmgr):
    # type: (ReportLocation, Step, EventManager) -> None

    thread_id = threading.current_thread().ident
    eventmgr.fire(events.StepStartEvent(location, step.description, thread_id, event_time=step.start_time))

    for log in step.get_logs():
        if isinstance(log, Log):
            eventmgr.fire(
                events.LogEvent(
                    location, step.description, thread_id, log.level, log.message, log.time
                )
            )
        elif isinstance(log, Attachment):
            eventmgr.fire(
                events.LogAttachmentEvent(
                    location, step.description, thread_id, log.filename, log.description,
                    log.as_image, log.time
                )
            )
        elif isinstance(log, Url):
            eventmgr.fire(
                events.LogUrlEvent(
                    location, step.description, thread_id, log.url, log.description, log.time
                )
            )
        elif isinstance(log, Check):
            eventmgr.fire(
                events.CheckEvent(
                    location, step.description, thread_id, log.description, log.is_successful,
                    log.details, log.time
                )
            )
        else:
            raise ValueError("Unknown step log %s" % log)

    eventmgr.fire(events.StepEndEvent(location, step.description, thread_id, event_time=step.end_time))


def _replay_steps_events(location, steps, eventmgr):
    # type: (ReportLocation, Iterable[Step], EventManager) -> None
    for step in steps:
        _replay_step(location, step, eventmgr)


def _replay_test_events(test, eventmgr):
    # type: (TestResult, EventManager) -> None
    if test.status in ("passed", "failed", None):  # None means "in progress"
        eventmgr.fire(events.TestStartEvent(test, test.start_time))
        _replay_steps_events(ReportLocation.in_test(test), test.get_steps(), eventmgr)
        if test.end_time:
            eventmgr.fire(events.TestEndEvent(test, test.end_time))
    elif test.status == "skipped":
        eventmgr.fire(events.TestSkippedEvent(test, test.status_details, test.start_time))
    elif test.status == "disabled":
        eventmgr.fire(events.TestDisabledEvent(test, test.status_details, test.start_time))
    else:
        raise ValueError("Unknown test status '%s'" % test.status)


def _replay_suite_events(suite, eventmgr):
    # type: (SuiteResult, EventManager) -> None

    eventmgr.fire(events.SuiteStartEvent(suite, suite.start_time))

    if suite.suite_setup:
        eventmgr.fire(events.SuiteSetupStartEvent(suite, suite.suite_setup.start_time))
        _replay_steps_events(ReportLocation.in_suite_setup(suite), suite.suite_setup.get_steps(), eventmgr)
        if suite.suite_setup.end_time:
            eventmgr.fire(events.SuiteSetupEndEvent(suite, suite.suite_setup.end_time))

    for test in suite.get_tests():
        _replay_test_events(test, eventmgr)

    for sub_suite in suite.get_suites():
        _replay_suite_events(sub_suite, eventmgr)

    if suite.suite_teardown:
        eventmgr.fire(events.SuiteTeardownStartEvent(suite, suite.suite_teardown.start_time))
        _replay_steps_events(ReportLocation.in_suite_teardown(suite), suite.suite_teardown.get_steps(), eventmgr)
        if suite.suite_teardown.end_time:
            eventmgr.fire(events.SuiteTeardownEndEvent(suite, suite.suite_teardown.end_time))

    if suite.end_time:
        eventmgr.fire(events.SuiteEndEvent(suite, suite.end_time))


def replay_report_events(report, eventmgr):
    # type: (Report, EventManager) -> None

    eventmgr.fire(events.TestSessionStartEvent(report, report.start_time))

    if report.test_session_setup:
        eventmgr.fire(events.TestSessionSetupStartEvent(report.test_session_setup.start_time))
        _replay_steps_events(ReportLocation.in_test_session_setup(), report.test_session_setup.get_steps(), eventmgr)
        if report.test_session_setup.end_time:
            eventmgr.fire(events.TestSessionSetupEndEvent(report.test_session_setup.end_time))

    for suite in report.get_suites():
        _replay_suite_events(suite, eventmgr)

    if report.test_session_teardown:
        eventmgr.fire(events.TestSessionTeardownStartEvent(report.test_session_teardown.start_time))
        _replay_steps_events(ReportLocation.in_test_session_teardown(), report.test_session_teardown.get_steps(), eventmgr)
        if report.test_session_teardown.end_time:
            eventmgr.fire(events.TestSessionTeardownEndEvent(report.test_session_teardown.end_time))

    if report.end_time:
        eventmgr.fire(events.TestSessionEndEvent(report, report.end_time))



'''
Created on Jan 4, 2017

@author: nicolas
'''

import os

from lemoncheesecake.cli.command import Command
from lemoncheesecake.project import create_project
from lemoncheesecake.exceptions import LemoncheesecakeException


class BootstrapCommand(Command):
    def get_name(self):
        return "bootstrap"

    def get_description(self):
        return "Create a new lemoncheesecake project"

    def add_cli_args(self, cli_parser):
        cli_parser.add_argument("dir", help="The directory where the project must be created")

    def run_cmd(self, cli_args):
        try:
            os.mkdir(cli_args.dir)
        except (IOError, OSError) as e:
            raise LemoncheesecakeException("Cannot create project directory '%s': %s" % (cli_args.dir, e))

        try:
            create_project(cli_args.dir)
        except (IOError, OSError) as e:
            raise LemoncheesecakeException("Cannot create project '%s': %s" % (cli_args.dir, e))

        return 0



#!/usr/bin/python

'''
Created on Aug 20, 2016

@author: nicolas
'''

from setuptools import setup, find_packages

import lemoncheesecake


setup(
    name="lemoncheesecake",
    version=lemoncheesecake.__version__,
    description="Test Storytelling",
    long_description=open("README.rst").read(),
    author="Nicolas Delon",
    author_email="nicolas.delon@gmail.com",
    license="Apache License (Version 2.0)",
    url="http://lemoncheesecake.io",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    keywords="QA tests functional",
    project_urls={
        'Documentation': 'http://docs.lemoncheesecake.io/',
        'Source': 'https://github.com/lemoncheesecake/lemoncheesecake',
        'Tracker': 'https://github.com/lemoncheesecake/lemoncheesecake/issues',
    },

    packages=find_packages(),
    include_package_data=True,
    install_requires=("colorama", "termcolor", "terminaltables", "six", "typing", "python-slugify"),
    extras_require={
        "xml": "lxml",
        "junit": "lxml",
        "reportportal": "reportportal-client~=3.0",
        "slack": "slacker"
    },
    entry_points={
        "console_scripts": [
            "lcc = lemoncheesecake.cli:main",
        ]
    }
)



'''
Created on Feb 14, 2017

@author: nicolas
'''

from __future__ import print_function
from collections import defaultdict

from lemoncheesecake.helpers.console import print_table, bold
from lemoncheesecake.cli.command import Command
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.project import load_project


class FixturesCommand(Command):
    def get_name(self):
        return "fixtures"

    def get_description(self):
        return "Show the fixtures available in the project"

    @staticmethod
    def show_fixtures(scope, fixtures, used_by_tests, used_by_fixtures):
        lines = []
        ordered_fixtures = sorted(
            fixtures,
            key=lambda f: used_by_fixtures.get(f.name, 0) + used_by_tests.get(f.name, 0),
            reverse=True
        )
        for fixt in ordered_fixtures:
            lines.append([
                bold(fixt.name), ", ".join(fixt.params or "-"),
                used_by_fixtures.get(fixt.name, 0), used_by_tests.get(fixt.name, 0)
            ])
        print_table(
            "Fixture with scope %s" % bold(scope),
            ["Fixture", "Dependencies", "Used by fixtures", "Used by tests"], lines
        )

    def run_cmd(self, cli_args):
        project = load_project()
        suites = project.load_suites()
        fixtures = project.load_fixtures()

        fixtures_by_scope = defaultdict(list)
        for fixt in fixtures:
            fixtures_by_scope[fixt.scope].append(fixt)

        used_by_tests = {}
        for test in flatten_tests(suites):
            for fixt_name in test.get_fixtures():
                used_by_tests[fixt_name] = used_by_tests.get(fixt_name, 0) + 1

        used_by_fixtures = {}
        for fixt in fixtures:
            for param in fixt.params:
                used_by_fixtures[param] = used_by_fixtures.get(param, 0) + 1

        for scope in "pre_run", "session", "suite", "test":
            self.show_fixtures(scope, fixtures_by_scope.get(scope, []), used_by_tests, used_by_fixtures)
            print()

        return 0



'''
Created on Mar 12, 2017

@author: nicolas
'''


class Command(object):
    def get_name(self):
        pass

    def get_description(self):
        return None

    def add_cli_args(self, cli_parser):
        pass

    def run_cmd(self, cli_args):
        pass



'''
Created on Apr 1, 2017

@author: nicolas
'''

from typing import Sequence, Any

from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.composites import is_


class EntryMatcher(object):
    def build_description(self):
        raise NotImplementedError()

    def get_entry(self, actual):
        # type: (dict) -> Any
        """Return the value of dict corresponding to entry matching or raise KeyError if entry is not found"""
        raise NotImplementedError()


class KeyPathMatcher(EntryMatcher):
    """Dict lookup through a list of key, each key represent a level of depth of the dict"""
    def __init__(self, path):
        # type: (Sequence[str]) -> None
        self.path = path

    def build_description(self):
        return " -> ".join(map(jsonify, self.path))

    def get_entry(self, actual):
        d = actual
        for key in self.path:
            try:
                d = d[key]
            # make sure to always raise a KeyError even if d is a list/tuple or something
            # that does not support __getitem__
            except (TypeError, IndexError):
                raise KeyError()
        return d


def wrap_key_matcher(key_matcher):
    if isinstance(key_matcher, EntryMatcher):
        return key_matcher
    if type(key_matcher) not in (list, tuple):
        key_matcher = (key_matcher,)
    return KeyPathMatcher(tuple(key_matcher))


class HasEntry(Matcher):
    def __init__(self, key_matcher, value_matcher):
        # type: (EntryMatcher, Matcher) -> None
        self.key_matcher = key_matcher
        self.value_matcher = value_matcher

    def build_short_description(self, transformation):
        ret = transformation('to have entry %s' % self.key_matcher.build_description())
        if self.value_matcher:
            ret += " that " + self.value_matcher.build_short_description(MatcherDescriptionTransformer(conjugate=True))
        return ret

    def build_description(self, transformation):
        ret = transformation('to have entry %s' % self.key_matcher.build_description())
        if self.value_matcher:
            ret += " that " + self.value_matcher.build_description(MatcherDescriptionTransformer(conjugate=True))
        return ret

    def matches(self, actual):
        try:
            value = self.key_matcher.get_entry(actual)
        except KeyError:
            return MatchResult.failure('No entry %s' % self.key_matcher.build_description())

        if self.value_matcher:
            return self.value_matcher.matches(value)
        else:
            return MatchResult.success("got %s" % jsonify(value))


def has_entry(key_matcher, value_matcher=None):
    """
    Test if dict has a <key> entry whose value matches (optional) value_matcher.
    Key entry can a standard dict key or a list of key where each element represent a
    level of depth of the dict (when dict are imbricated)
    """
    return HasEntry(
        wrap_key_matcher(key_matcher),
        is_(value_matcher) if value_matcher is not None else None
    )



'''
Created on Mar 12, 2017

@author: nicolas
'''

from __future__ import print_function

import os.path as osp
import sys
import platform

import lemoncheesecake
from lemoncheesecake.project import load_project
from lemoncheesecake.reporting import get_reporting_backends
from lemoncheesecake.reporting.reportdir import DEFAULT_REPORT_DIR_NAME
from lemoncheesecake.testtree import filter_suites
from lemoncheesecake.exceptions import UserError, ProjectNotFound

LEMONCHEESECAKE_VERSION = "lemoncheesecake version %s (using Python %s - %s)" % (
    lemoncheesecake.__version__, platform.python_version(), sys.executable
)


def load_suites_from_project(project, test_filter=None):
    suites = project.load_suites()
    if all(suite.is_empty() for suite in suites):
        raise UserError("No test is defined in your lemoncheesecake project.")

    project.metadata_policy.check_suites_compliance(suites)

    if test_filter:
        suites = filter_suites(suites, test_filter)
        if len(suites) == 0:
            raise UserError("The filter does not match any test")

    return suites


def auto_detect_reporting_backends():
    try:
        project = load_project()
        return project.reporting_backends.values()
    except ProjectNotFound:
        return get_reporting_backends()


def add_report_path_cli_arg(cli_parser):
    cli_parser.add_argument("report_path", nargs='?', help="Report file or directory")


def get_report_path(cli_args):
    # first attempt: has the report path been specified on the CLI ?
    if cli_args.report_path:
        return cli_args.report_path

    # second attempt: is there a report on the current working directory ?
    if osp.exists(DEFAULT_REPORT_DIR_NAME):
        return DEFAULT_REPORT_DIR_NAME

    # third attempt: try to find a project and then the corresponding report
    try:
        project = load_project()
    except ProjectNotFound:
        pass
    else:
        report_path = osp.join(project.dir, DEFAULT_REPORT_DIR_NAME)
        if osp.exists(report_path):
            return report_path

    # could not find anything
    raise UserError("Cannot find report path")



from __future__ import absolute_import
import threading


class ThreadedFactory(object):
    """
    .. versionadded:: 1.9.0

    This factory class returns a new object per thread when calling :py:func:`get_object`.

    This class works by subclassing and:

    - implementing the :py:func:`setup_object` is mandatory
    - implementing the :py:func:`teardown_object` is optional

    NB: if the ``__init__`` method is overridden then the base class ``__init__`` method
    must be called.
    """

    def __init__(self):
        self._local = threading.local()
        self._objects = []

    def get_object(self):
        """
        Get the object, it will be only created once per thread.

        :return: object
        """
        try:
            return self._local.object
        except AttributeError:
            obj = self.setup_object()
            self._local.object = obj
            self._objects.append(obj)
            return obj

    def teardown_factory(self):
        """
        Teardown the factory.

        This method must be called if :py:func:`teardown_object` has been implemented.
        """
        for obj in self._objects:
            self.teardown_object(obj)

    def setup_object(self):
        """
        Create the object. This method MUST be implemented.

        :return: object
        """
        raise NotImplementedError()

    def teardown_object(self, obj):
        """
        Teardown an object. You can implement this method if your object needs a special teardown phase.

        :param obj: an object created by the :py:func:`setup_object`
        """
        pass
